"""
The functions in this module should provide an excellent starting place for learning about how
math inspector works, and what makes it different from other programming environments
"""
"""
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import numpy as np
from skimage import measure as _measure
from skimage.color import hsv2rgb as _hsv2rgb

NEG_Z = -1e-12
POS_Z = 1e-12

def complex_grid(x_min=-10, x_max=10, y_min=-10, y_max=10, num=25):
	"""
creates a grid of intersecting lines made of complex numbers

Parameters
----------
x_min : float
    the minimum x-coordinate
x_max : float
    the maximum x-coordinate
y_min : float
	the minimum y-coordinate
y_max : float
    the maximum y-coordinate
num : int
	the number of grid lines per axis


Notes
-----
If the grid passes through either the x or y axis, then this function returns a grid for each quadrant of the plane occupied by the grid.  This is important when using complex_grid as input to a ufunc to avoid visual artificats that result from points which are very close together but have a different sign.

Examples
--------

To create a complex grid, we will use the numpy ``meshgrid`` method, the first step is to create two arrays of evenly spaced numbers

>>> from numpy import linspace, meshgrid, transpose
>>> x = linspace(0,10)
>>> y = linspace(0,10)

We can then create a set of lines `Z` in the following way

>>> Y,X = meshgrid(y,x)
>>> Z = X + 1j*Y

To plot the grid, we transponse `Z` to get the lines to go in the other direction as well

>>> plot(Z,transpose(Z))
    """
	if (x_min < 0 and x_max > 0) or (y_min < 0 and y_max > 0):
		return (*complex_grid(x_min, NEG_Z, y_min, NEG_Z, int(num/2)),
			*complex_grid(POS_Z, x_max, y_min, NEG_Z, int(num/2)),
			*complex_grid(x_min, NEG_Z, POS_Z, y_max, int(num/2)),
			*complex_grid(POS_Z, x_max, POS_Z, y_max, int(num/2)))

	x = np.linspace(x_min, x_max, num+1)
	y = np.linspace(y_min, y_max, num+1)
	X, Y = np.meshgrid(x,y)
	Z = X + 1j*Y

	return Z, np.transpose(Z)

def power(points, exponent):
	"""
draws the result of taking the power of a set of points to an arbitrary exponent

Parameters
----------
points : array
    a numpy array

exponent: float
	the exponent


Examples
--------

>>> from mathinspector.examples import complex_grid
>>> from numpy import power
>>> t=1
>>> plot(power(complex_grid(), 2))

Animation:

>>> app.animate("power", argname="x2", start=2, stop=-2, step=-0.005, delay=0.01)

    """
	return points**exponent

def times_table(factor, start=0, end=199):
	"""
creates a visual representation of the remainders of a number modulo a factor f,
by representing these numbers as points along a circle

Parameters
----------
factor : float
    the minimum x-coordinate
start : int
    the maximum x-coordinate
end : int
	the minimum y-coordinate

Notes
-----
If the grid passes through either the x or y axis, then this function returns a grid for each quadrant of the plane occupied by the grid.  This is important when using complex_grid as input to a ufunc to avoid visual artificats that result from points which are very close together but have a different sign.

References
----------
Mathologer, Times Tables, Mandelbrot and the Heart of Mathematics,
https://www.youtube.com/watch?v=qhbuKbxJsk8

Examples
--------
>>> from mathinspector.examples import times_table
>>> t=2
>>> plot(times_table(t))

Animation

>>> app.animate("t", start=2, stop=50)
    """

	result = []

	for value in range(start, end):
		new_value = (value * factor) % end
		result.append([
			(np.cos(value * 2 * np.pi / end), np.sin(value * 2 * np.pi / end)),
			(np.cos(new_value * 2 * np.pi / end), np.sin(new_value * 2 * np.pi / end))
		])

	return tuple(result);

def elliptic_curve(q,p):
	"""
draws an elliptic curve with parameters q,p

:math: y^3 = x^3 + p*x + q

Parameters
----------
p : float
    coefficient of x

q : float
    constant term

Notes
-----
This function is implemented by using the library skimage to find the set of points (x,y) which satisfy the equation y^2 = x^3 + p*x + q


The polynomial form of an elliptic curve is related to the differential equation satisfied by the Weierstrass elliptic function[1], by making the substitution

ro(z)′ = y

ro(z)  = x

References
----------
[0] https://en.wikipedia.org/wiki/Elliptic_curve


[1] https://en.wikipedia.org/wiki/Weierstrass%27s_elliptic_functions#Differential_equation

Examples
--------

>>> from mathinspector.examples import elliptic_curve
>>> plot(elliptic_curve(0,0))
    """
	x, y = np.ogrid[-12:12:100j, -12:12:100j]
	r = y**2 - x**3 - x*p - q

	return _measure.find_contours(r, 0)

def four_leaf_rose(q=4):
	"""
draws the algebraic variety known as the "four leaf rose"

:math: (x^2 + y^2)^3 - q*(x^2)*(y^2) = 0

Parameters
----------
q : float
    the coefficient of the cross term

Notes
-----
This function is implemented by using the library skimage to find the set of points (x,y) which satisfy the equation (x^2 + y^2)^3 - q*(x^2)*(y^2) = 0

Examples
--------

>>> from mathinspector.examples import four_leaf_rose
>>> plot(four_leaf_rose())
    """
	x, y = np.ogrid[-12:12:100j, -12:12:100j]
	r = (x**2 + y**2)**3 - q*(x**2)*(y**2)
	return _measure.find_contours(r, 0)


def transform(f, g, X, t):
	"""
animate the transition between two functions

:math: w(f,g,X,t) = f(X)*(t-1) + g(X)*t

Parameters
----------
f : array
    the first function to blend
g : array
    the second function to blend
X : array
	the range where the functions f and g are evaluated
t : float
    A number between 0 and 1 which determines the amount of blending between f and g


Examples
--------
>>> from mathinspector.examples import transform
>>> from numpy import exp, sqrt, linspace
>>> transform(exp, sqrt, linspace(-3,3), 0.5)
[ 0.5        -0.69980788  0.62129323 -0.29568233 -0.14752521]

Graphical illustration:

>>> from mathinspector.examples import complex_grid
>>> t = 0.5
>>> plot(transform(exp, sqrt, complex_grid(), t))

Animated Transition:

>>> app.animate("t", start=1, stop=0, step=-0.001)
    """
	return f(X)*(t-1) + g(X)*t

def complex_points(position, size, step):
	"""
creates a matrix of complex points using numpy methods

:math: (x^2 + y^2)^3 - q*(x^2)*(y^2) = 0

Parameters
----------
position : tuple
    the (x,y) position of the center of the points

size : tuple
    the (width,height) of the square which contains the point set

step : float
    the spacing between points

Notes
-----
This function uses the numpy `arange` and `meshgrid` functions to create a set of complex points.  View the source code for this example to see the implementation details.

Examples
--------

>>> from mathinspector.examples import complex_points
>>> complex_points((0,0),(2,2),0.5)
[[-1. +1.j  -1. +0.5j -1. +0.j  -1. -0.5j]
 [-0.5+1.j  -0.5+0.5j -0.5+0.j  -0.5-0.5j]
 [ 0. +1.j   0. +0.5j  0. +0.j   0. -0.5j]
 [ 0.5+1.j   0.5+0.5j  0.5+0.j   0.5-0.5j]]
    """
	x0, y0 = position
	w, h = size
	y = np.arange(y0 + h/2, y0 - h/2, -step)
	x = np.arange(x0 - w/2, x0 + w/2, step)
	Y, X = np.meshgrid(y,x)
	return X + 1j*Y

def domain_coloring(position, size, step, transform=None):
	"""
Domain coloring is a popular technique for visualizing the output of complex valued functions of a
complex variable.

Parameters
----------
position : tuple
    the (x,y) position of the center of the pixels

size : tuple
    the (width,height) of the square which contains the pixel set

step : float
    the distance between adjacent pixels

Examples
--------
>>> from mathinspector.examples import domain_coloring
>>> plot(pixelmap=domain_coloring)

References
----------
3blue1brown, Winding numbers and domain coloring, https://www.youtube.com/watch?v=b7FxPsqfkOY


https://en.wikipedia.org/wiki/Domain_coloring


	"""
	Z = complex_points(position, size, step)
	if transform:
		Z = transform(Z)
	H = np.angle(Z) % (2*np.pi)
	r = np.log2(1. + np.abs(Z))
	S = (1. + np.abs(np.sin(2. * np.pi * r))) / 2.
	V = (1. + np.abs(np.cos(2. * np.pi * r))) / 2.

	color_wheel = np.empty((Z.shape[0], Z.shape[1], 3), dtype=np.uint8)
	color_wheel[:, :, 0] = 255 * (H / H.max())
	color_wheel[:, :, 1] = 255 * S
	color_wheel[:, :, 2] = 255 * V

	return 255 * _hsv2rgb(color_wheel)

def mandelbrot_set(position, size, step, iterations=80):
	"""
The Mandlebrot set is one of the most famous fractals, this function computes it's defining equation up to a given number of iterations.

:math: $M = lim_{n \\to iterations} z_n < 2$ where $z_{n+1}=z_n^2 + z_0$

Parameters
----------
position : tuple
    the (x,y) position of the center of the pixels

size : tuple
    the (width,height) of the square which contains the pixel set

step : float
    the distance between adjacent pixels

iterations : int
	the number of iterations before bailing out of the calculation

References
----------
https://en.wikipedia.org/wiki/Mandelbrot_set


https://tomroelandts.com/articles/how-to-compute-the-mandelbrot-set-using-numpy-array-operations

Examples
--------
>>> from mathinspector.examples import mandelbrot_set
>>> plot(pixelmap=mandelbrot_set)

	"""
	C = complex_points(position, size, step)
	C_size = C.shape[0], C.shape[1]
	Z = np.zeros(C_size, dtype=complex)
	M = np.full(C_size, True, dtype=bool)

	for i in range(iterations):
		Z[M] = Z[M] * Z[M] + C[M]
		M[np.abs(Z) > 2] = False

	M = np.uint8(M)
	R = np.empty((C_size[0], C_size[1], 3), dtype=np.uint8)
	R[:,:,0] = 255 * M
	R[:,:,1] = 255 * M
	R[:,:,2] = 255 * M
	return R

def helix(t, as_points=False):
	"""
draws a three dimensional helix

:math: helix(t) = (np.cos(t), t, np.sin(t))

Parameters
----------
t : float
    parameter of the curve

Examples
--------
>>> from mathinspector.examples import helix
>>> from numpy import linspace
>>> T = 10
>>> plot(helix(linspace(0,T)))

Animated:

>>> app.animate("T", start=10, stop=25)
    """
	result = np.cos(t), t, np.sin(t)
	return result if as_points else list(result)

def wireframe(x,y,fn):
	"""
draws the wireframe line for a surface given by a function of two real variables f(x,y)

:math: wireframe(t) = (x, y, fn(x,y))

Parameters
----------
x  : array
	the x values of the range

y  : array
	the y values of the range

fn : callable
    a real valued function of two real variables

Examples
--------
>>> from numpy import linspace
>>> from mathinspector.examples import wireframe
>>> X = linspace(0,10)
>>> Y = linspace(0,10)
>>> plot(wireframe(X,Y, lambda x,y: (x,y,x**2 + y**2)))
    """

	x_lines, y_lines = [], []
	for xval in x:
		temp = []
		for yval in y:
			temp.append(fn(xval,yval))
		x_lines.append(temp)

	for yval in y:
		temp = []
		for xval in x:
			temp.append(fn(xval,yval))
		y_lines.append(temp)

	return x_lines, y_lines

def surface(x,y,fn):
	"""
draws quads for a surface given by a function of two real variables f(x,y)

:math: surface(t) = (x, y, fn(x,y))

Parameters
----------
x  : array
	the x values of the range

y  : array
	the y values of the range

fn : callable
    a real valued function of two real variables

Examples
--------
>>> from mathinspector.examples import surface
>>> X = linspace(0,10)
>>> Y = linspace(0,10)
>>> plot(surface(X,Y, lambda x,y: (x,y,x**2 + y**2)))
	"""

	return wireframe(x,y,fn)[1]

def cylinder(pos=(0,0,0),radius=1,num=25, as_wireframe=False):
	"""
draws a cylinder at a given position with a given radius and number of subdivisions

:math: f(t) = (cos(t), sin(t), r)

Parameters
----------
pos  : tuple
	position of the center of the cylinder

radius : float
	radius of the cylinder

num : int
    number of subdivisions

Examples
--------
>>> from mathinspector.examples import cylinder
>>> plot(cylinder())
    """
	result = wireframe(
		np.linspace(-radius,radius, num=num),
		np.linspace(0, 2*np.pi, num=num),
		lambda r,t: (pos[0] + np.cos(t), pos[1] + np.sin(t), pos[2] + r)
	)

	return result if as_wireframe else result[1]

def sphere(pos=(0,0,0),radius=1,num=12, as_wireframe=False):
	"""
draws a sphere at a given position with a given radius and number of subdivisions

:math: f(t) = (cos(t), sin(t), r)

Parameters
----------
pos  : tuple
	position of the center of the cylinder

radius : float
	radius of the cylinder

num : int
    number of subdivisions

Notes
-----
There are some known issues with the way this function has been implemented related to how the normals are calculated by the OpenGLWindow of the `plot` module.

Examples
--------
>>> from mathinspector.examples import sphere
>>> plot(sphere())
    """
	result = wireframe(
		np.linspace(-np.pi,np.pi, num=num),
		np.linspace(-np.pi,np.pi, num=num),
		lambda p,t: (pos[0] + radius*np.cos(t)*np.cos(p), pos[1] + radius*np.cos(t)*np.sin(p), pos[2] + radius*np.sin(t))
	)
	return result if as_wireframe else result[1]




"""
Math Inspector: a visual programming environment for scientific computing
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import pygame, glm, ctypes, os
import numpy as np
from pygame.locals import *
from OpenGL.GL import *
from OpenGL.GL import shaders
from OpenGL.GLU import *
from .util import is_iterable, instanceof
from .shader import Shader
from pygame._sdl2.video import Window


OPTIONS = {
	"size": (1280,768),
	"title": "Math Inspector",
	"on_update": None,
	"on_close": None,
	"show_grid": True,
}

window = None

class OpenGLWindow:
	def __init__(self):
		self.zoom = 1;
		self.model = glm.mat4(1.0)
		self.camera_pos = glm.vec3(0,10,7)
		self.camera_front = glm.normalize(glm.vec3(0,-1,-1))
		self.camera_up = glm.vec3(0,1,0)

		self.view = glm.lookAt(self.camera_pos, self.camera_pos * self.camera_front, self.camera_up)
		self.projection = glm.perspective(glm.radians(45), OPTIONS["size"][0]/OPTIONS["size"][1], 0.1, 1000)
		self.model = glm.rotate(self.model, glm.radians(-90.0), glm.vec3(0,1,0))
		self.light_pos = glm.vec3(10,10,10)
		self.is_running = False
		self.is_animation_running = False
		# self.points = []
		# self.lines = []
		# self.quads = []
		# self.line_offset = []
		# self.quad_offset = []

	def plot(self, *args, **kwargs):
		OPTIONS.update(kwargs)
		self.projection = glm.perspective(glm.radians(45), OPTIONS["size"][0]/OPTIONS["size"][1], 0.1, 1000)
		pygame.init()
		pygame.display.gl_set_attribute(GL_CONTEXT_MAJOR_VERSION, 4)
		pygame.display.gl_set_attribute(GL_CONTEXT_MINOR_VERSION, 1)
		pygame.display.gl_set_attribute(pygame.GL_CONTEXT_PROFILE_MASK, pygame.GL_CONTEXT_PROFILE_CORE)
		self.screen = pygame.display.set_mode(OPTIONS["size"], DOUBLEBUF | OPENGL | RESIZABLE | HWSURFACE)

		glClearColor(39/255,40/255,44/255,1)
		glEnable(GL_BLEND)
		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA)
		glEnable(GL_DEPTH_TEST)
		# glEnable(GL_MULTISAMPLE);
		# glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST)

		self.generate_vertices(*args)
		pygame.display.set_caption(OPTIONS["title"])

		last_tick = 0
		animation_timer = 0
		did_change = True
		self.is_running = True
		is_focused = False
		keypress = pygame.key.get_pressed()

		if OPTIONS["on_update"]:
			OPTIONS["on_update"]()

		while self.is_running:
			tick = pygame.time.get_ticks()/1000
			delta_time = tick - last_tick
			last_tick = tick
			speed = 10 * delta_time
			camera_right = glm.normalize(glm.cross(self.camera_front, self.camera_up))

			for event in pygame.event.get():
				if event.type == MOUSEWHEEL:
					delta = 1 + event.y/100
					self.zoom *= delta
					self.model = glm.scale(self.model, glm.vec3(delta))
					did_change = True
				elif event.type == MOUSEMOTION:
					if event.buttons[0]:
						fx,fy,fz = self.camera_front
						rotate_front = glm.rotate(glm.mat4(1), glm.radians(-event.rel[0]/10), self.camera_up)
						rotate_up = glm.rotate(glm.mat4(1), glm.radians(-event.rel[1]/10), camera_right)
						self.camera_front = (rotate_front * glm.vec4(fx,fy,fz,1)).xyz
						ux,uy,uz = self.camera_front
						self.camera_front = (rotate_up * glm.vec4(ux,uy,uz,1)).xyz
						did_change = True
					elif event.buttons[2]:
						self.camera_pos -= speed * event.rel[0] * camera_right
						self.camera_pos += speed * event.rel[1] * self.camera_up
						did_change = True
				elif event.type == pygame.USEREVENT:
					if hasattr(event, "args"):
						if event.args:
							self.generate_vertices(*event.args)
					if hasattr(event, "kwargs"):
						OPTIONS.update(event.kwargs)
					elif hasattr(event, "animate"):
						delay = event.animate[0]
						callback = event.animate[1]
						self.is_animation_running = True
					did_change = True
				elif event.type == VIDEORESIZE:
					did_change = True
					OPTIONS["size"] = event.size
					self.projection = glm.perspective(glm.radians(45), event.size[0]/event.size[1], 0.1, 1000)
				elif event.type == ACTIVEEVENT:
					is_focused = bool(event.gain)
				elif (event.type == pygame.QUIT
					or (event.type == KEYUP and event.key == K_ESCAPE)
					or (event.type == KEYUP and event.key == K_w and keypress[K_LMETA])
				):
					self.is_running = False
					if OPTIONS["on_close"]:
						OPTIONS["on_close"]()

			if self.is_animation_running:
				if animation_timer >= delay:
					animation_timer = 0
					values = callback()
					if values:
						self.generate_vertices(*values)
						did_change = True
					else:
						self.is_animation_running = None
				else:
					animation_timer += delta_time

			if not is_focused and OPTIONS["on_update"]:
				OPTIONS["on_update"]()

			keypress = pygame.key.get_pressed()
			if is_focused:
				if keypress[K_w] or keypress[K_UP]:
					self.camera_pos += speed * self.camera_front
					did_change = True
				if keypress[K_s] or keypress[K_DOWN]:
					self.camera_pos -= speed * self.camera_front
					did_change = True
				if keypress[K_d] or keypress[K_RIGHT]:
					self.camera_pos += speed * camera_right
					did_change = True
				if keypress[K_a] or keypress[K_LEFT]:
					self.camera_pos -= speed * camera_right
					did_change = True

			if did_change:
				self.view = glm.lookAt(self.camera_pos, self.camera_pos + self.camera_front, self.camera_up)
				mvp = self.projection * self.view * self.model

				glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

				if len(self.points) > 0:
					glUseProgram(self.shader["points"].program)
					glUniformMatrix4fv(self.shader["points"]["mvp"], 1, GL_FALSE, glm.value_ptr(mvp))
					glBindVertexArray(self.point_vao)
					glDrawArrays(GL_POINTS, 0, int(len(self.points)/3))

				if len(self.lines) > 0:
					glUseProgram(self.shader["lines"].program)
					glUniformMatrix4fv(self.shader["lines"]["mvp"], 1, GL_FALSE, glm.value_ptr(mvp))
					glBindVertexArray(self.line_vao)
					for i in range(0, len(self.line_offset)):
						glDrawArrays(GL_LINE_STRIP, *self.line_offset[i])

				if len(self.quads) > 0:
					glUseProgram(self.shader["quads"].program)
					glUniformMatrix4fv(self.shader["quads"]["model"], 1, GL_FALSE, glm.value_ptr(self.model))
					glUniformMatrix4fv(self.shader["quads"]["view"], 1, GL_FALSE, glm.value_ptr(self.view))
					glUniformMatrix4fv(self.shader["quads"]["projection"], 1, GL_FALSE, glm.value_ptr(self.projection))
					glUniform3f(self.shader["quads"]["lightPos"], *self.light_pos.xyz)
					glBindVertexArray(self.quad_vao)
					for i in range(0, len(self.quad_offset)):
						glDrawArrays(GL_TRIANGLE_STRIP, *self.quad_offset[i])

				if OPTIONS["show_grid"]:
					glUseProgram(self.shader["grid"].program)
					glUniformMatrix4fv(self.shader["grid"]["model"], 1, GL_FALSE, glm.value_ptr(self.model))
					glUniformMatrix4fv(self.shader["grid"]["view"], 1, GL_FALSE, glm.value_ptr(self.view))
					glUniformMatrix4fv(self.shader["grid"]["proj"], 1, GL_FALSE, glm.value_ptr(self.projection))
					glUniform1f(self.shader["grid"]["zoom"], self.zoom)
					glUniform3f(self.shader["grid"]["cameraPos"], *self.camera_pos.xyz)
					glBindVertexArray(self.grid_vao)
					glDrawArrays(GL_TRIANGLE_STRIP, 0, 4)

				pygame.display.flip()
				did_change = False

		win_w, win_h = Window.from_display_module().position
		OPTIONS["window_pos"] = str(win_w) + ", " + str(win_h)
		os.environ["SDL_VIDEO_WINDOW_POS"] = OPTIONS["window_pos"]
		pygame.display.quit()

	def generate_vertices(self, *args):
		# TODO - run the loop over args here, that way i can use recursion in self.get_vertices easily
		self.points, self.lines, self.quads, self.line_offset, self.quad_offset = self.get_vertices(*args)

		self.grid = np.array([-1,-1, 0,
							-1, 1, 0,
							 1,-1, 0,
							 1, 1, 0 ], dtype='float32')

		self.grid_vao = glGenVertexArrays(1)
		glBindVertexArray(self.grid_vao)
		self.grid_vbo = glGenBuffers(1)
		glBindBuffer(GL_ARRAY_BUFFER, self.grid_vbo)
		glBufferData(GL_ARRAY_BUFFER, self.grid, GL_STATIC_DRAW)
		glEnableVertexAttribArray(0)
		glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, ctypes.c_void_p(0))
		self.shader = { "grid": Shader("grid_vert.glsl", "grid_frag.glsl", ("view", "proj", "model", "zoom", "cameraPos")) }

		if len(self.points) > 0:
			self.point_vao = glGenVertexArrays(1)
			glBindVertexArray(self.point_vao)
			self.point_vbo = glGenBuffers(1)
			glBindBuffer(GL_ARRAY_BUFFER, self.point_vbo)
			glBufferData(GL_ARRAY_BUFFER, self.points, GL_STATIC_DRAW)
			glEnableVertexAttribArray(0)
			glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, None)
			self.shader["points"] = Shader("point_vert.glsl", "point_frag.glsl", "mvp")

		if len(self.lines) > 0:
			self.line_vao = glGenVertexArrays(1)
			glBindVertexArray(self.line_vao)
			self.line_vbo = glGenBuffers(1)
			glBindBuffer(GL_ARRAY_BUFFER, self.line_vbo)
			glBufferData(GL_ARRAY_BUFFER, self.lines, GL_STATIC_DRAW)
			glEnableVertexAttribArray(0)
			glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, None)
			self.shader["lines"] = Shader("line_vert.glsl", "line_frag.glsl", "mvp")

		if len(self.quads) > 0:
			self.quad_vao = glGenVertexArrays(1)
			glBindVertexArray(self.quad_vao)
			self.quad_vbo = glGenBuffers(1)
			glBindBuffer(GL_ARRAY_BUFFER, self.quad_vbo)
			glBufferData(GL_ARRAY_BUFFER, self.quads, GL_STATIC_DRAW)
			glEnableVertexAttribArray(0)
			glEnableVertexAttribArray(1)
			glEnableVertexAttribArray(2)
			glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 8*4, ctypes.c_void_p(0))
			glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, 8*4, ctypes.c_void_p(3*4))
			glVertexAttribPointer(2, 3, GL_FLOAT, GL_FALSE, 8*4, ctypes.c_void_p(5*4))
			self.shader["quads"] = Shader("quad_vert.glsl", "quad_frag.glsl", ("model", "view", "projection", "lightPos"))

	def get_vertices(self, *args, line_num=0, quad_num=0):
		points, lines, quads, line_offset, quad_offset = [], [], [], [], []
		for arg in args:
			if instanceof(arg, list) and instanceof(arg[0], list) and instanceof(arg[0][0], (int,float)):
				line_offset.append((line_num, len(arg[0])))
				line_num += len(arg[0])
				for i in range(0, len(arg[0])):
					lines.extend([arg[0][i], arg[1][i], arg[2][i]])
			elif instanceof(arg, tuple):
				if len(arg) == 2 and instanceof(arg[0], list):
					for line in arg[0]:
						line_offset.append((line_num, len(line)))
						line_num += len(line)
						for vert in line:
							lines.extend(list(vert))

					for line in arg[1]:
						line_offset.append((line_num, len(line)))
						line_num += len(line)
						for vert in line:
							lines.extend(list(vert))
				elif instanceof(arg[0], list) and len(arg) == 3 and len(arg[0]) == len(arg[1]) == len(arg[2]):
					for i in range(0, len(arg[0])):
						points.extend([arg[0][i], arg[1][i], arg[2][i]])
				elif len(arg) == 3 and instanceof(arg[0], (int,float)):
					points.extend(list(arg))
				else:
					temp = self.get_vertices(*arg, line_num=line_num, quad_num=quad_num)
					points.extend(temp[0])
					lines.extend(temp[1])
					quads.extend(temp[2])
					line_offset.extend(temp[3])
					quad_offset.extend(temp[4])
			elif instanceof(arg, list):
				if instanceof(arg[0], tuple):
					line_offset.append((line_num, len(arg)))
					line_num += len(arg)
					for j in arg:
						lines.extend(list(j))
				elif instanceof(arg[0], list):
					quad_start = quad_num
					for i in range(0, len(arg)-1):
						for j in range(0, len(arg[i])-1):
							a = glm.vec3(*arg[i][j+1]) - glm.vec3(*arg[i][j])
							b = glm.vec3(*arg[i+1][j]) - glm.vec3(*arg[i][j])
							normal = glm.normalize(glm.cross(b,a))
							x_len = len(arg) - 1
							y_len = len(arg[j]) - 1
							quads.extend(list(arg[i][j]) + [i/x_len, j/y_len] + list(normal))
							quads.extend(list(arg[i][j+1]) + [i/x_len, (j+1)/y_len] + list(normal))
							quads.extend(list(arg[i+1][j]) + [(i+1)/x_len, j/y_len] + list(normal))
							quads.extend(list(arg[i+1][j+1]) + [(i+1)/x_len, (j+1)/y_len] + list(normal))
							quad_num += 4
					quad_offset.append((quad_start, quad_num))

		return (
			np.array(points, dtype="float32"),
			np.array(lines, dtype="float32"),
			np.array(quads, dtype="float32"),
			line_offset,
			quad_offset
		)

	def config(self, **kwargs):
		pygame.event.post(pygame.event.Event(pygame.USEREVENT, kwargs=kwargs))

	def update(self, *args, **kwargs):
		pygame.event.post(pygame.event.Event(pygame.USEREVENT, args=args, kwargs=kwargs))

	def animate(self, delay, callback):
		pygame.event.post(pygame.event.Event(pygame.USEREVENT, animate=(delay,callback)))

	def close(self):
		pygame.event.post(pygame.event.Event(pygame.QUIT))





"""
Math Inspector: a visual programming environment for scientific computing
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import inspect, sys, traceback, os, re
import numpy as np
import tkinter as tk
from code import InteractiveInterpreter
from types import CodeType

from .. import plot
from ..doc import Help
from ..style import Color, TAGS
from ..util import vdict, open_editor, BUTTON_RIGHT, BASEPATH, FONT
from ..widget import Text, Menu
from .builtin_print import builtin_print
from .codeparser import CodeParser
from .prompt import Prompt, FONTSIZE
try:
	from ..version import VERSION
except ImportError:
	VERSION = "UNKNOWN"

RE_TRACEBACK = r"^()(Traceback \(most recent call last\))"
RE_FILEPATH = r"(File (\"(?!<).*\"))"
RE_INPUT = r"(File \"(<.*>)\")"
RE_LINE = r"((line [0-9]*))"
RE_IN = r"line ([0-9]*), in (.*)"

class Interpreter(Text, InteractiveInterpreter):
	"""
	This class extends `InteractiveInterpreter` from the module `code`,
	and follows the official patterns for creating python interpreters.

	In order to synchronize all the different views with the interpreter, the local namespace
	is stored in a vdict called `self.locals`.  A vdict is key-value pair, just like a regular
	dictionary, except it also has callbacks for events such as setting and deleting items.

	When running commands, `self.locals` is passed to the builtin `exec` function as the third parameter.
	This parameter determines the local namespace exec runs in.  You can display the current
	contents of the local namespace by running the command

	>>> locals()

	When exec processses code, it updates the values in `self.locals`, but since this is a vdict,
	the callbacks are automatically triggered.

	The math inspector console extends the traditional python interpreter, and has many quality of
	life improvements such as syntax highlighting and a wide variety of hotkeys. Before and
	after each command is executed, the command string is parsed to keep the other views
	synchronized with the variables in the local namespace.  Code parsing is done
	using the abstract syntax tree module `ast`.  An abstract syntax tree is generated by
	python under the hood as an intermediate step between processing
	lines of code and executing those commands.

	An excellent resource for learning about abstract syntax tree's is the
	website https://greentreesnakes.readthedocs.io/
	"""
	def __init__(self, app):
		InteractiveInterpreter.__init__(self, vdict({
			"app": app,
			"plot": plot.plot
		}, setitem=self.setitem, delitem=self.delitem))

		self.frame = tk.Frame(app,
			padx=16,
			pady=8,
			background=Color.DARK_BLACK)

		Text.__init__(self, self.frame,
			readonly=True,
			background=Color.DARK_BLACK,
			font=FONT,
			padx=0,
			pady=0,
			wrap="word",
			cursor="arrow",
			insertbackground=Color.DARK_BLACK)

		sys.displayhook = self.write
		sys.excepthook = self.showtraceback
		__builtins__["print"] = self.write
		__builtins__["help"] = Help(app)
		__builtins__["clear"] = Clear(self)
		__builtins__["copyright"] = "Copyright (c) 2018-2021 Matt Calhoun.\nAll Rights Reserved."
		__builtins__["credits"] = "Created by Matt Calhoun.\nSee https://mathinspector.com for more information."
		__builtins__["license"] = License()

		self.app = app
		self.prompt = Prompt(self, self.frame)
		self.parse = CodeParser(app)
		self.buffer = []
		self.prevent_module_import = False

		self.bind("<Key>", self._on_key_log)
		self.bind("<Configure>", self.prompt.on_configure_log)
		self.bind("<ButtonRelease-1>", self.on_click_log)
		self.pack(fill="both", expand=True)

		for i in ["error_file"]:
			self.tag_bind(i, "<Motion>", lambda event, key=i: self._motion(event, key))
			self.tag_bind(i, "<Leave>", lambda event, key=i: self._leave(event, key))
			self.tag_bind(i, "<Button-1>", lambda event, key=i: self._click(event, key))

	def _init(self, event):
		self.config_count = self.config_count + 1 if hasattr(self, "config_count") else 1
		if self.config_count > 2:
			self.do_greet()
			self.prompt.place(width=self.winfo_width())
			self.bind("<Configure>", self.prompt.on_configure_log)

	def do_greet(self):
		self.write("Math Inspector " + VERSION + " (Beta)\nType \"help\", \"copyright\", \"credits\", \"license\" for more information")
		self.prompt()

	def setitem(self, key, value):
		if inspect.ismodule(value):
			if not self.prevent_module_import:
				self.app.modules[key] = value
		else:
			self.app.objects[key] = value

	def delitem(self, key, value):
		if inspect.ismodule(value):
			del self.app.modules[key]
		else:
			del self.app.objects[key]

	def synclocals(self):
		self.locals.store.update(self.app.objects.store)
		self.locals.store.update(self.app.modules.store)

	def eval(self, source):
		self.synclocals()
		return eval(source, self.locals)

	def exec(self, source, filename="<file>"):
		self.prevent_module_import = True
		self.synclocals()
		self.parse.preprocess(source)
		self.runsource(source, filename, "exec")
		self.parse.postprocess(source)
		self.prevent_module_import = False

	def push(self, s, filename="<input>", log=True, symbol="single"):
		self.synclocals()
		source = "".join(self.buffer + [s])
		self.parse.preprocess(source)
		if source[:4] == "plot":
			self.prompt()
		did_compile = self.runsource(source, filename, symbol)

		if did_compile:
			self.buffer.append(s + "\n")
			self.prompt.history.append(s)
		else:
			self.parse.postprocess(source)
			self.prompt.history.extend(s)
			self.buffer.clear()
		self.prompt()

	def write(self, *args, syntax_highlight=False, tags=(), **kwargs):
		idx = self.index("insert")
		for r in args:
			if re.match(RE_TRACEBACK, str(r)):
				tags = ("red", *tags)

			if r is not None:
				if isinstance(r, Exception):
					builtin_print("\a")
					tags = tuple(list(tags) + ["red"])
				self.insert("end", str(r), tags, syntax_highlight=syntax_highlight)
				if len(args) > 1:
					self.insert("end", "\t")

		try:
			self.highlight(RE_FILEPATH, "error_file", idx)
			self.highlight(RE_INPUT, "purple", idx)
			self.highlight(RE_LINE, "blue", idx)
			self.highlight(RE_IN, "green", idx)
		except:
			pass

		if self.get("1.0", "end").strip():
			self.insert("end", "\n")
		self.see("end")
		self.prompt.move()

	def showtraceback(self, *args):
		sys.last_type, sys.last_value, last_tb = ei = sys.exc_info()
		sys.last_traceback = last_tb

		try:
			lines = traceback.format_exception(ei[0], ei[1], last_tb.tb_next)
			self.write(''.join(lines).rstrip(), tags="red")
			builtin_print ("\a")
			self.app.menu.setview("console", True)
		finally:
			last_tb = ei = None

	def clear(self):
		self.prompt.is_on_bottom = False
		self.delete("1.0", "end")

	def _on_key_log(self, event):
		result = self._on_key(event)
		if result:
			self.prompt.focus()
		return result

	def on_click_log(self, event):
		 if not self.tag_ranges("sel"):
		 	self.prompt.focus()

	def _on_button_right(self, event):
		option = []
		tag_ranges = self.tag_ranges("sel")
		if tag_ranges:
			option.extend([{
				"label": "Copy",
				"command": lambda: self.clipboard_append(self.get(*tag_ranges))
			}])

		self.menu.show(event, option + [{
			"label": "Clear Log",
			"command": clear
		}])


	def _click(self, event, tag):
		if not self.hover_range: return
		content = self.get(*self.hover_range)
		if tag == "error_file":
			open_editor(os.path.abspath(content[1:-1]))

class License:
	def __call__(self):
		help(os.path.join(BASEPATH, "LICENSE"))

	def __repr__(self):
		return """
Math Inspector: a visual programming environment for scientific computing

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Type license() to see the full license text
		"""

class Clear:
	def __init__(self, console):
		self.console = console

	def __call__(self):
		self.console.clear()
		self.console.prompt()

	def __repr__(self):
		self.console.clear()
		self.console.prompt()
		return ""



"""
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import inspect, re, string
from .docscrape import FunctionDoc

RE_FN = re.compile(r"[a-zA-Z_0-9]*\((.*)\)")
RE_EXTRA = re.compile(r"(\[.*\])")

def argspec(obj, withself=True):
    """
    tries to return inspect.getargspec, as a backup tries to parse the FunctionDoc
    """
    if not callable(obj):
        return [], {}

    if isinstance(obj, dict):
        return [], {}

    try:
        fullargspec = inspect.getfullargspec(obj)
    except:
        fullargspec = None

    try:
        doc = FunctionDoc(obj)
    except:
        doc = None

    if fullargspec is not None:
        num_kwargs = len(fullargspec[3]) if fullargspec[3] else 0
        args = fullargspec[0][0 if withself else 1:len(fullargspec[0]) - num_kwargs]
        num_args = len(args) if withself else 1 + len(args)
        kwargs = {}
        if num_args + num_kwargs > 0:
            if num_kwargs > 0:
                kwargs = { fullargspec[0][i]:fullargspec[3][i - num_args] for i in range(num_args, len(fullargspec[0])) }
            if fullargspec[5] and len(fullargspec[5]) > 0:
                for name in fullargspec[5]:
                    kwargs[name] = fullargspec[5][name]

            return args, kwargs

    if doc is not None:
        signature = doc["Signature"].replace(", /", "").replace(", \*", "").replace(", *", "")
        signature = RE_EXTRA.sub("", signature)
        params = RE_FN.findall(signature)
        if len(params) == 0:
            return False
        items = [i.lstrip() for i in params[0].split(",")]

        args = [i for i in items if "=" not in i]
        kwargs = {}
        for i in items:
            if "=" in i:
                key, val = i.split("=")
                try:
                    val = eval(val)
                except:
                    val = None
                kwargs[key] = val

        return args, kwargs

    return ['x'], {}


"""
Math Inspector: a visual programming environment for scientific computing
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from pkg_resources import resource_filename
import inspect, os, platform, sys, pkg_resources, builtins, keyword, math
from sys import builtin_module_names
from pkgutil import iter_modules

SYSTEM = platform.system()
BUTTON_RIGHT = '<Button-3>' if SYSTEM in ("Windows", "Linux") else '<Button-2>'
BUTTON_RELEASE_RIGHT = '<ButtonRelease-3>' if SYSTEM in ("Windows", "Linux") else '<ButtonRelease-2>'
BUTTON_RIGHT_MOTION = '<B3-Motion>' if SYSTEM in ("Windows", "Linux") else '<B2-Motion>'
CONTROL_KEY = 'Control' if SYSTEM in ("Windows", "Linux") else 'Command'
if hasattr(sys, "_MEIPASS"):
    BASEPATH = os.path.join(
        sys._MEIPASS,
        "assets" if SYSTEM in ("Windows", "Linux") else "../assets")
    if SYSTEM == "Windows":
        AUTOSAVE_PATH = os.path.join(os.path.join(os.getenv('LOCALAPPDATA'), "MathInspector"), "autosave.math")
    else:
        AUTOSAVE_PATH = os.path.join(BASEPATH, "autosave.math")
else:
    BASEPATH = resource_filename("mathinspector", "assets")
    AUTOSAVE_PATH = os.path.join(BASEPATH, "autosave.math")

MESSAGE_TIMEOUT = 4000

INSTALLED_PKGS = sorted([str(i).split(" ")[0] for i in pkg_resources.working_set], key=str.casefold)
BUILTIN_PKGS = sorted([j.name for j in iter_modules()], key=str.casefold)

BUILTIN_CLASS = [i for i in dir(builtins) if inspect.isclass(getattr(builtins, i))]
BUILTIN_FUNCTION = [i for i in dir(builtins) if callable(getattr(builtins, i)) and i not in BUILTIN_CLASS]
BUILTIN_CONSTANT = [i for i in dir(builtins) if i not in BUILTIN_FUNCTION and i not in BUILTIN_CLASS]
BUILTIN_MODULES = builtin_module_names
KEYWORD_LIST = keyword.kwlist

ZOOM_IN = 1.1
ZOOM_OUT = 0.9
HITBOX = 32
FONTSIZE = "12" # TODO - refactor this into the rest of the FONTSIZE stuff
PROMPT_FONTSIZE = 18.5

if SYSTEM in ("Windows", "Linux"):
    FONT = "LucidaConsole 12"
    DOC_FONT = "Nunito-ExtraLight 12"
    FONT_SIZE = {
        "extra-small": "8",
        "small": "10",
        "default": "12",
        "medium": "16",
        "large": "18",
        "extra-large": "22",
    }
    ITEM_FONTSIZE = {
        "name": "14",
        "class": "8",
        "value": "12",
        "argname": "8",
        "argvalue": "10",
    }
else:
    FONT = "Menlo 15"
    DOC_FONT = "Nunito-ExtraLight 16"
    FONT_SIZE = {
        "extra-small": "12",
        "small": "14",
        "default": "15",
        "medium": "18",
        "large": "22",
        "extra-large": "24",
    }
    ITEM_FONTSIZE = {
        "name": "18",
        "class": "12",
        "value": "16",
        "argname": "8",
        "argvalue": "12",
    }

EXCLUDED_MODULES = [
    "absolute_import",
    "division",
    "print_function",
    "testing",
    "version",
    "add_newdoc",
    "add_newdocs",
    "add_docstring",
    "add_newdoc_ufunc"
]

# EXCLUDED_SUBMODULES = INSTALLED_PKGS + BUILTIN_PKGS

EXCLUDED_DIR = [
    "__pycache__",
    ".git"
]

EXCLUDED_FILES = [
    ".DS_Store",
    "__pycache__",
]

EXCLUDED_EXT = [
    ".pyc",
    ".DS_Store",
    ".math"
]



"""
Math Inspector: a visual programming environment for scientific computing
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import tkinter as tk
from tkinter import ttk
import inspect
from ..util.argspec import argspec
from ..util.common import classname
from ..util.config import BUTTON_RIGHT, EXCLUDED_MODULES, INSTALLED_PKGS, BUILTIN_PKGS
from ..console.builtin_print import builtin_print
from ..widget import Notebook, Treeview, Button, Menu
from .show_textfile import show_textfile
from .doc import Doc
from ..style import Color

class Browser(tk.Toplevel):

    def __init__(self, app, obj=None, title=None, browse=False, on_import=None, geometry=None):
        tk.Toplevel.__init__(self, app, background=Color.BLACK)
        if geometry:
            self.geometry(geometry)

        self.protocol("WM_DELETE_WINDOW", self._on_close)

        if obj:
            self.title(title or classname(obj))
            self.doc = Doc(self, obj, run_code=app.console.prompt.push)
            self.doc.pack(fill="both", expand=True)
            return

        self.app = app
        self.on_import = on_import
        self.menu = Menu(self)

        self.paned_window = ttk.PanedWindow(self, orient="horizontal")
        self.notebook = Notebook(self, has_labels=True)
        self.installed_pkgs = Treeview(self, self)
        self.builtin_pkgs = Treeview(self, self)
        self.notebook.add("Built In", self.builtin_pkgs)
        self.notebook.add("Installed", self.installed_pkgs)
        self.doc = Doc(self, has_sidebar=True, run_code=app.console.prompt.push)

        frame = tk.Frame(self)
        input_area = tk.Frame(frame, background=Color.BLACK)
        name_label = tk.Label(input_area, text="import ", font="Menlo 14", background=Color.BLACK, foreground=Color.WHITE)
        self.name_entry = tk.Entry(input_area, font="Monospace 14", highlightbackground=Color.BLACK)

        input_area_2 = tk.Frame(frame, background=Color.BLACK)
        as_label = tk.Label(input_area_2, text="    as ", font="Menlo 14", background=Color.BLACK, foreground=Color.WHITE)
        self.as_entry = tk.Entry(input_area_2, font="Monospace 14", highlightbackground=Color.BLACK)

        btn_area = tk.Frame(frame, background=Color.BLACK)
        cancel_btn = Button(btn_area, text="Cancel", command=self.destroy)
        self.ok_btn = Button(btn_area, text="Ok", command=self._on_ok)

        for i in INSTALLED_PKGS:
            self.installed_pkgs.insert("", "end", i, text=str(i).split(" ")[0])

        for j in BUILTIN_PKGS:
            self.builtin_pkgs.insert("", "end", j, text=j)

        if on_import:
            self.notebook.frame.pack(fill="both", expand=True)
            frame.pack(side="left", fill="both", expand=True)
            self.name_entry.focus()
        else:
            self.paned_window.add(self.notebook.frame)
            self.paned_window.add(self.doc)
            self.paned_window.pack(side="left", fill="both", expand=True)

        if on_import:
            cancel_btn.pack(side="left", fill="x", expand=True, anchor="center", padx=4, pady=4)
            self.ok_btn.pack(side="left", fill="x", expand=True, anchor="center", padx=4, pady=4)
            btn_area.pack(side="bottom", fill="both")

            input_area_2.pack(side="bottom", fill="both")
            as_label.pack(side="left")
            self.as_entry.pack(side="left", expand=True)

            input_area.pack(side="bottom", fill="both")
            name_label.pack(side="left")
            self.name_entry.pack(side="left", expand=True)

            self.bind("<Return>", self._on_ok)
            self.bind("<Escape>", lambda event: self.destroy())
        else:
            self.doc.paned_window.bind("<Configure>", self._on_configure_doc)

        for i in ("installed_pkgs", "builtin_pkgs"):
            getattr(self, i).bind("<<TreeviewSelect>>", lambda event, key=i: self._on_selection(event, key))
            getattr(self, i).bind(BUTTON_RIGHT, lambda event, key=i: self._on_button_right(event, key))

    def _on_configure_doc(self, event):
        self.doc.paned_window.sashpos(0,0)
        self.doc.paned_window.unbind("<Configure>")

    def _on_selection(self, event, name):
        key = getattr(self, name).selection()[0]

        self.doc.clear()
        if self.on_import:
            self.name_entry.delete(0, "end")
            self.name_entry.insert("end", key)
            return

        try:
            module = __import__(key)
        except:
            return

        self.doc.show(module)

    def _on_close(self):
        help.geometry = self.geometry()
        self.destroy()

    def _on_ok(self, event=None):
        name = self.name_entry.get()
        if not name:
            builtin_print("\a")
            return

        if self.on_import:
            self.on_import(self.name_entry.get(), self.as_entry.get())
        self.destroy()

    def _on_button_right(self, event, name):
        attr = getattr(self, name)
        key = attr.identify_row(event.y)
        attr.selection_set(key)

        if self.on_import:
            try:
                module = __import__(key)
            except:
                return

            items = [{
                "label": "View Doc",
                "command": lambda: help(module)
            }]
        else:
            items = [{
                "label": "Import Module",
                "command": lambda: self.app.menu.import_module(key, open_folders=True)
            }]

        self.menu.set_menu(items)
        self.menu.show(event)



"""
Math Inspector: a visual programming environment for scientific computing
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import tkinter as tk
from ..util import vdict
from ..console.builtin_print import builtin_print


class Menu(tk.Menu):
	def __init__(self, parent, items=[], **kwargs):
		tk.Menu.__init__(self, parent, tearoff=False)
		self._ = {}
		self.parent = parent
		self.set_menu(items)

	def set_menu(self, items, prev=None):
		if not prev:
			self.delete(0, 100)
			prev = self

		for j in items:
			# builtin_print(j)
			if "menu" in j:
				if len(j["menu"]) > 0:
					menu = Menu(self.parent)
					self._[j["label"]] = menu
					self.set_menu(j["menu"], menu)
					kwargs = j.copy()
					kwargs["menu"] = menu
					prev.add_cascade(**kwargs)
			elif "separator" in j:
				prev.add_separator(j["separator"])
			elif "label" in j:
				prev.add_command(**j)

				if "accelerator" in j:
					self.parent.bind("<" + j["accelerator"].replace("+", "-") + ">", j["command"])

	def show(self, event, items=None):
		if items:
			self.set_menu(items)
		self.post(event.x_root, event.y_root)



"""
this class adds extra functionality to the builtin help function,
and is used to override the the builtin help function by calling

>>> __builtins__["help"] = Help(app)

in console/interpreter.py

"""
"""
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import builtins
from ..util.config import INSTALLED_PKGS, BUILTIN_PKGS, BUILTIN_CLASS
from ..util.common import name_ext
from os import path
from ..console.builtin_print import builtin_print
from .browser import Browser
from . import manual

class Help:
	def __init__(self, app):
		self.app = app
		self.geometry = None
		self.browser = None

	def __call__(self, key=None, title=None):
		if not key:
			self.browser = Browser(self.app, manual, "mathinspector", geometry=self.geometry)
			return
		obj = self.getobj(key)
		if not obj: return
		self.browser = Browser(self.app, obj, title=title, geometry=self.geometry)

	def __repr__(self):
		return "Type help() for interactive help, help(object) for help about object, or help.browse() to view all available documentation."

	def browse(self, callback=None):
		self.browser = Browser(self.app, browse=True, on_import=callback)

	def import_module(self, callback):
		self.browser = Browser(self.app, on_import=callback)

	def getobj(self, key):
		if key is None: return manual

		if not isinstance(key, str):
			return key

		if key in INSTALLED_PKGS + BUILTIN_PKGS:
			try:
				obj = __import__(key)
				return obj
			except:
				pass

		if key in self.app.objects:
			if isinstance(self.app.objects[key], tuple([getattr(builtins,i) for i in BUILTIN_CLASS])):
				return self.app.objects[key].__class__
			return self.app.objects[key]
		if key in self.app.modules:
			return self.app.modules[key]
		if path.isfile(key):
			name, ext = name_ext(key)
			if name in self.app.modules:
				return self.app.modules[name]

			if name in ("LICENSE") or ext in (".md", ".rst"):
				return key
		try:
			module, attr = key.rsplit('.', 1)
			return getattr(self.app.modules[module], attr)
		except:
			try:
				obj, attr = key.split('.', 1)
				return getattr(self.app.objects[obj], attr)
			except:
				return None



"""
Math Inspector: a visual programming environment for scientific computing
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import tkinter as tk
from tkinter import ttk
from ..style import Color
from ..util.config import FONT_SIZE

class Notebook(ttk.Notebook):
	def __init__(self, *args, has_labels=False, **kwargs):
		self.frame = tk.Frame(*args, background=Color.BLACK, highlightthickness=4, highlightbackground=Color.BLACK, **kwargs)
		ttk.Notebook.__init__(self, self.frame)

		self._tabs = {}
		self.has_labels = has_labels

		if has_labels:
			self.nav = tk.Frame(self.frame, background=Color.BLACK, height=32)
			self.nav.pack(fill='x', expand=False)

		self.pack(fill='both', expand=True)
		self.bind('<<NotebookTabChanged>>', self._on_changetab)

	def add(self, name, widget, **kwargs):
		super(Notebook, self).add(widget, **kwargs)
		tab_id = super(Notebook, self).tabs()[-1]
		if self.has_labels:
			tab = NotebookTab(self, name)
			self._tabs[tab_id] = tab
			tab.label.bind("<Button-1>", lambda event: self.select(tab_id))

		return tab_id

	def _on_changetab(self, event):
		tab_id = self.select()
		if tab_id and self.has_labels:
			self._tabs[tab_id].select()
			selected = self.select()
			if selected and selected != tab_id:
				self._tabs[selected].unselect()

			for i in self._tabs:
				if i == tab_id:
					self.select(i)
				elif i in self._tabs:
					self._tabs[i].unselect()

class NotebookTab(tk.Frame):
	def __init__(self, parent, name):

		tk.Frame.__init__(self, parent.nav, background=Color.BLACK)
		self.label = tk.Label(self,
			text=name,
			font="SourceSansPro " + FONT_SIZE["extra-small"] + " bold",
			foreground=Color.WHITE,
			background=Color.BLACK,
			padx=12,
			pady=8
		)

		self.parent = parent
		self.name = name
		self.is_selected = False

		self.label.pack(side="left", fill="both", expand=True)
		self.pack(fill="x", expand=True, side="left")

		self.bind("<Enter>", self._on_enter)
		self.bind("<Leave>", self._on_leave)

	def select(self):
		self.is_selected = True
		self.label.config(background=Color.BLACK, foreground=Color.ORANGE)

	def unselect(self):
		self.is_selected = False
		self.label.config(background=Color.BLACK, foreground=Color.WHITE)

	def _on_enter(self, event):
		if not self.is_selected:
			self.label.config(background=Color.VERY_DARK_GREY)

	def _on_leave(self, event):
		if not self.is_selected:
			self.label.config(background=Color.BLACK)



"""
Based on: https://stackoverflow.com/questions/48567935/get-parameterarg-count-of-builtin-functions-in-python
"""
import re

def numargs(func):
    arg_test = 0
    while arg_test < 999:
        try:
            s = [None] * arg_test
            func(*s)
        except TypeError as e:
            message = str(e)
            found = re.match(
                r"[\w]+\(\) takes ([0-9]{1,3}) positional argument[s]* but " +
                str(arg_test) + " were given", message)

            if "takes no positional arguments" in message:
                return 0
            if found:
                return int(found.group(1))

            if "takes no arguments" in message:
                return 0
            elif "takes at most" in message:
                found = re.match(
                    r"[\w]+\(\) takes at most ([0-9]{1,3}).+", message)
                if found:
                    return int(found.group(1))
            elif "takes exactly" in message:
                # string can contain 'takes 1' or 'takes one',
                # depending on the Python version
                found = re.match(
                    r"[\w]+\(\) takes exactly ([0-9]{1,3}|[\w]+).+", message)
                if found:
                    return 1 if found.group(1) == "one" \
                            else int(found.group(1))
        arg_test += 1
    return -1  # *args



"""
Math Inspector: a visual programming environment for scientific computing
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import tkinter as tk
import textwrap, traceback
from pprint import pformat
from ..util import fontcolor
from ..style import Color
from ..widget.text import Text

class Entry(Text):
	def __init__(self, item):
		Text.__init__(self, item.canvas,
			foreground=Color.VERY_LIGHT_GREY,
			# background=Color.GREEN,
			background=Color.VERY_LIGHT_PURPLE,
			padx=0,
			pady=0)

		self.item = item
		self.canvas = item.canvas
		self.app = item.canvas.app
		self.argname = None
		self.window = None
		self.edit_id = None
		self.font = None
		self.tag_configure("center", justify="center")
		self.bind("<Key>", self._on_key)
		self.bind("<<Modify>>", self._on_edit_change)

	def edit(self, canvas_id, color=Color.VERY_LIGHT_GREY):
		self.edit_id = canvas_id
		self.argname = self.canvas.getname(canvas_id, "argvalue")
		self.canvas.itemconfig(canvas_id, text="")
		self.delete("1.0", "end")
		# REFACTOR - way too busy with these nested if's
		arg = None if self.argname == "<value>" else self.item.args[self.argname] if self.argname in self.item.args else self.item.kwargs[self.argname]
		self.insert("end", self.item.content(truncate=False) if self.argname == "<value>" else "" if arg is None else self.item.content(arg, truncate=False))

		font = self.canvas.itemconfig(canvas_id)["font"][4]
		params = font.split(" ")
		fontsize = 12 if len(params) < 2 else int(int(params[1]) / self.canvas.zoom)

		self.font = params[0] + " " + str(int(fontsize)) + ("" if len(params) < 3 else (" " + " ".join(params[2:])))
		if self.argname == "<value>":
			self.config(font=font, **self.item.text_dimensions(self.item.obj))
		else:
			self.config(font=font, **self.item.text_dimensions(arg))

		if isinstance(self.item.obj, str):
			self.config(foreground=Color.YELLOW)

		opts = { "window": self }

		self.window = self.canvas.create_window(*self.canvas.coords(canvas_id), **opts)
		self.focus()

	def _on_edit_change(self, event=None):
		if self.item.is_callable: return
		try:
			content = self.item.obj.__class__(self.get("1.0", "end"))
		except:
			content = self.get("1.0", "end")

		self.config(**self.item.text_dimensions(content, truncate=False))
		self.item.resize(content)

		if not isinstance(self.item.obj, str):
			self.syntax_highlight()

	def zoom(self):
		params = self.font.split(" ")
		fontsize = int(12 if len(params) < 2 else int(params[1]))
		self.config(font=params[0] + " " + str(int(fontsize * self.canvas.zoom)) + ("" if len(params) < 3 else (" " + params[2])))

	def finish(self, cancel=False):
		if cancel:
			if self.argname == "<value>":
				content = self.item.content()
			else:
				arg = self.item.args[self.argname] if self.argname in self.item.args else self.item.kwargs[self.argname]
				content = self.item.content(arg)

			self.canvas.itemconfig(self.edit_id, text=content)
			self.item.resize()
			self.hide()
			return

		content = self.get("1.0", "end-1c")
		if content == "":
			content = "None"

		if self.argname == "<value>":
			try:
				result = self.app.console.eval(content)
			except:
				result = content

			try:
				self.app.objects[self.item.name] = self.item.obj.__class__(result)
			except:
				self.app.console.prompt.push("")
				self.app.console.showtraceback()
				self.finish(cancel=True)
				return
		elif self.argname in self.item.args:
			self.item.args[self.argname] = self.app.console.eval(content)
		else:
			self.item.kwargs[self.argname] = self.app.console.eval(content)

		self.hide()

	def hide(self):
		self.canvas.delete(self.window)
		self.window = None
		self.edit_id = None
		self.font = None
		self.canvas.edit_item = False

	def _on_key(self, event):
		if event.keysym == 'Escape':
			self.finish(cancel=True)
			return "break"
		elif event.keysym == 'Return':
			self.finish()
			return "break"



"""
Math Inspector: a visual programming environment for scientific computing
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
RE_MARKDOWN = {
	"h1": r"^#(?!#) {0,}(.*)\n",
	"h2": (r"^##(?!#) {0,1}(.*)", r"^(?!\n)(.*\n===={0,})$", r"^(?!\n)(.*\n----{0,})$"),
	"h3": r"^###(?!#) {0,}(.*)",
	"italic": r"(?<!\*)\*(?!\*)(.*?)\*",
	"bold": r"\*\*(?!\*)(.*?)\*\*",
	"blockquote": r"^>(?!>)(.*)",
	"unordered_list": r"^((\t|    ) {0,}-(?!-).*)",
	"code": r"`(?!`)(.*?)`",
	"code_block": r"(?s)```\n{0,}(.*?)\n{0,}```.*$",
	"horizontal_rule": (r"^()----{0,}$", r"^()===={0,}$")
}

RE_DOC = {
    "bold": r"()``(.*?)``",
    "menlo_italic": r"()(?<!`)`(?!`)(.*?)`",
    "link_url": r"()(http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+~]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+)",
    "code_sample": r"()(?s)(>>>.*?\n)\n",
    "console_prompt": r"((>>>))"
}



"""
Math Inspector: a visual programming environment for scientific computing
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import inspect, traceback
from .. import plot
from ..style import Color
from ..util import vdict
from ..util import classname, fontcolor, argspec
from ..util.config import ITEM_FONTSIZE as FONTSIZE
from .entry import Entry
from textwrap import fill
from pprint import pformat
from numpy import ufunc

FONT_SIZE = 20
MAX_NUMBER_WIDTH = 10
MAX_STR_WIDTH = 256
TEXTWRAP_SIZE = 28
NODE_SIZE = 6

class Item:
    def __init__(self, canvas, name, coord=(0,0), args={}, kwargs={}, connection=None, opts={}, class_name=None, is_output_item=False):
        self.canvas = canvas
        self.app = canvas.app
        self.name = name
        self.obj = self.app.objects[name]
        self.connection = connection
        self.argspec = argspec(self.obj)#, withself=inspect.isclass(self.obj))

        self.args = vdict({},
            getitem=lambda key: self.getarg(key, "args"),
            setitem=self.setarg)

        self.kwargs = vdict({},
            getitem=lambda key: self.getarg(key, "kwargs"),
            setitem=self.setarg)

        self.opts = vdict({
            "show_kwargs": opts["show_kwargs"] if "show_kwargs" in opts else False,
            "sticky_graph": opts["sticky_graph"]  if "sticky_graph" in opts else  False,
            "line_color": opts["line_color"]  if "line_color" in opts else  Color.BLUE
        }, setitem=self.option)

        self.entry = Entry(self)
        self.is_callable = callable(self.obj)
        self.classname = class_name or classname(self.obj)
        self.show_wire = False
        self.method = None
        self.cached_value = None
        self.should_cache_value = True

        self.ids = {}
        x, y = coord
        width, height = self.dimensions(self.obj)
        node = NODE_SIZE * self.canvas.zoom
        create_parent = getattr(self.canvas, "create_" + ("rectangle" if callable(self.obj) or isinstance(self.obj, (str, list, dict)) else "oval"))

        self.ids["parent"] = create_parent(x - width/2, y - height/2, x + width/2, y + height/2,
            tags=("parent", "draggable", "name=" + self.name),
            outline=Color.INACTIVE,
            fill=Color.BLACK)

        self.ids["name"] = canvas.create_text(0,0,
            tags=("name", "draggable", "name=" + self.name, "font", "fontsize="+FONTSIZE["name"], "stop_scaling"),
            text=name or '',
            fill=Color.WHITE,
            font="Menlo 18")

        self.ids["class"] = canvas.create_text(0,0,
            tags=("class", "draggable", "name=" + self.name, "font", "fontsize="+FONTSIZE["argvalue"], "stop_scaling"),
            width=0,
            text=self.classname,
            fill=Color.BLUE)

        self.ids["output"] = canvas.create_oval(0,0,0,0,
            tags=("output", "name=" + self.name),
            outline=Color.INACTIVE,
            fill=Color.EMPTY_NODE)

        self.ids["wire"] = canvas.create_line(0,0,0,0,
            tags=("wire", "name=" + self.name, "scale_width"),
            fill=Color.INACTIVE,
            smooth=True,
            state="hidden",
            width=1)

        if not self.is_callable:
            self.ids["value"] = canvas.create_text(x,y,
                tags=("value", "draggable", "editable", "name=" + self.name, "argvalue=<value>", "font", "fontsize="+FONTSIZE["value"]),
                text=self.content(),
                fill=fontcolor(self.obj),
                font="Menlo 16")

            self.args.store["<value>"] = args["<value>"] if "<value>" in args else self.obj
            self.ids["arg=<value>"] = canvas.create_oval(0,0,0,0,
                tags=("input", "arg=<value>", "name=" + self.name),
                outline=Color.INACTIVE,
                fill=Color.EMPTY_NODE)
        else:
            num_args = len(self.argspec[0]) if not self.opts["show_kwargs"] else (len(self.argspec[0]) + len(self.argspec[1]))
            offset = ((2 * num_args) or 1 - height/2) * self.canvas.zoom
            i = 0 if num_args == 1 else (1 - num_args + 0.2)
            for j in self.argspec[0]:
                if j not in self.args:
                    self.args.store[j] = args[j] if j in args else None
                self.ids["arg=" + j] = canvas.create_oval(0,0,0,0,
                    tags=("input", "arg=" + j, "name=" + self.name),
                    outline=Color.INACTIVE,
                    fill=Color.EMPTY_NODE)

                self.ids["argname=" + j] = canvas.create_text(0,0,
                    tags=("name=" + self.name, "draggable", "font", "fontsize="+FONTSIZE["argname"]),
                    text=j,
                    fill=Color.DARK_ORANGE,
                    anchor="w")

                self.ids["argvalue=" + j] = canvas.create_text(0,0,
                    tags=("argvalue=" + j, "draggable", "editable", "name=" + self.name, "font", "fontsize="+FONTSIZE["argvalue"]),
                    text="None",
                    fill=Color.RED,
                    anchor="w")
                i += 2

            for k in self.argspec[1]:
                if k not in self.kwargs:
                    self.kwargs.store[k] = kwargs[k] if k in kwargs else self.argspec[1][k]
                self.ids["arg=" + k] = canvas.create_oval(0,0,0,0,
                    tags=("arg=" + k, "input", "name=" + self.name),
                    state="normal" if self.opts["show_kwargs"] else "hidden",
                    outline=Color.INACTIVE,
                    fill=Color.EMPTY_NODE)

                self.ids["argname=" + k] = canvas.create_text(0,0,
                    text=k,
                    tags=("name=" + self.name, "argname="+k, "draggable", "font", "fontsize="+FONTSIZE["argname"]),
                    state="normal" if self.opts["show_kwargs"] else "hidden",
                    fill=Color.ORANGE,
                    anchor="w")

                self.ids["argvalue=" + k] = canvas.create_text(0,0,
                    text=str(self.kwargs[k]),
                    tags=("argvalue=" + k, "draggable", "editable", "name=" + self.name, "font", "fontsize="+FONTSIZE["argvalue"]),
                    state="normal" if self.opts["show_kwargs"] else "hidden",
                    fill=fontcolor(self.kwargs[k]),
                    anchor="w")
                i += 2

        self.resize()
        if connection:
            output_name, output_arg = connection
            output = self.canvas[output_name]
            if output_arg in output.args:
                output.args[output_arg] = self
            else:
                output.kwargs[output_arg] = self
            self.move_wire()

        if is_output_item:
            self.app.node.output.connect(self)

    # REFACTOR - not crazy about this function having two purposes, should use .setvalue instead
    def value(self, obj=None):
        if obj is None:
            if self.is_callable and self.hasargs():
                try:
                    return self.obj(*[self.args[i] for i in self.args], **{ k: self.kwargs[k] for k in self.kwargs if self.kwargs.store[k] is not None })
                except Exception as err:
                    pass
            return self.obj

        # if "<value>" in self.args["connection"]:
        #     self.args["connection"]["<value>"].disconnect()

        self.obj = obj
        if not self.is_callable:
            self.config("value", text=self.content())

        if self.connection:
            name, argname = self.connection
            if argname in self.canvas[name].args:
                self.canvas[name].args[argname] = self
            elif argname in self.canvas[name].kwargs:
                self.canvas[name].kwargs[argname] = self

        if self in self.canvas.output.items:
            self.canvas.output.update_value(self)

        self.move_wire()
        self.resize()

    def getarg(self, key, attr):
        params = getattr(self, attr).store
        if key == "connection":
            return { i: params[i] for i in params if isinstance(params[i], Item) }
        return params[key].value() if isinstance(params[key], Item) else params[key]

    def setarg(self, key, value):
        prev = self.args.store[key] if key in self.args["connection"] else self.kwargs.store[key] if key in self.kwargs["connection"] else None
        result = value
        if isinstance(value, Item):
            if value == self: return False

            if key == "<value>":
                try:
                    result = self.obj.__class__(value.value())
                except Exception as err:
                    result = value.value()
            else:
                result = value.value()
            value.connection  = self.name, key
            self.config("arg="+key, fill=Color.PURPLE)
            value.config("output", fill=Color.PURPLE)
            value.move_wire()

            x, y = self.position()
            x1, y1 = value.position()
            offset = self.canvas.zoom * 160
            if x1 + offset > x:
                value.move(-(x1 - x + offset), 0)
                for i in value.args["connection"]:
                    value.args["connection"][i].move(-(x1 - x + offset), 0)
                for i in value.kwargs["connection"]:
                    value.kwargs["connection"][i].move(-(x1 - x + offset), 0)

        if prev:
            if value is None:
                if key in self.args:
                    self.args.store[key] = self.args.store[key].value()
                elif key in self.kwargs:
                    self.kwargs.store[key] = self.kwargs.store[key].value()
                prev.disconnect()
                return False
            elif prev != value:
                prev.disconnect()

        if key == "<value>":
            self.app.objects[self.name] = result
        else:
            self.config("argvalue="+key, text=str(result)[:MAX_NUMBER_WIDTH], fill=fontcolor(result))

        self.app.objects.item(self.name + "<arg=" + key + ">",
            text=str(result),
            tags=("editable", fontcolor(result, as_string=True)))

        if self.connection:
            name, argname = self.connection
            item = self.canvas[name]
            if argname in item.args:
                item.args[argname] = self
            elif argname in item.kwargs:
                item.kwargs[argname] = self

        if self in self.canvas.output.items:
            if key in self.args:
                self.args.store[key] = value
            elif key in self.kwargs:
                self.kwargs.store[key] = value
            self.canvas.output.update_value(self)
            return False


    def disconnect(self):
        if self.connection:
            name, argname = self.connection
            self.connection = None
            item = self.canvas[name]
            if argname in item.args:
                item.args[argname] = None
            elif argname in item.kwargs:
                item.kwargs[argname] = None
            item.config("arg="+argname, fill=Color.EMPTY_NODE)
            self.config("output", fill=Color.EMPTY_NODE)
        self.move_wire()

    def option(self, key, value=None):
        value = value if value is not None else not self.opts[key]
        self.opts.store[key] = value
        if key == "line_color":
            color = str(value) or Color.BLUE
        elif key == "show_kwargs":
            for j in self.kwargs:
                for k in ("arg="+j, "argname="+j, "argvalue="+j):
                    self.config(k, state="normal" if self.opts["show_kwargs"] else "hidden")
            self.resize()
            self.move()
        return False

    def content(self, *args, truncate=True):
        obj = self.obj if not args else args[0]
        if callable(obj): return str(self.value())

        if isinstance(obj, dict):
            return pformat(obj, indent=4, width=MAX_NUMBER_WIDTH)

        result = fill(str(obj), width=TEXTWRAP_SIZE) if isinstance(obj, (str, list)) else str(obj)
        if not truncate:
            return result

        width = MAX_STR_WIDTH if isinstance(obj, (str, list)) else MAX_NUMBER_WIDTH
        return result if len(result) < width else result[:width] + "..."

    def resize(self, content=None):
        content = content if content is not None else self.content()
        try:
            x, y = self.position()
        except:
            return
        width, height = self.dimensions(content)
        node = NODE_SIZE * self.canvas.zoom

        self.coords("parent", x - width/2, y - height/2, x + width/2, y + height/2)
        self.coords("output", x + width/2 - node, y - node, x + width/2 + node, y + node)
        self.coords("name", x, y - height/2 - 20 * self.canvas.zoom)
        # self.canvas.scale_font()
        self.coords("class", x, y + height/2 + 20 * self.canvas.zoom)
        self.move_wire()

        num_args = len(self.argspec[0]) if not self.opts["show_kwargs"] else len(self.argspec[0]) + len(self.argspec[1])
        num_args = num_args or 1
        offset = 2 * num_args
        i = 0 if num_args == 1 else (1 - num_args + 0.2)
        for j in self.args.store:
            self.coords("arg="+j, x - width/2 - node, y - node + i*height/offset, x - width/2 + node, y + node + i*height/offset)
            self.coords("argname="+j, x - width/2 + 12*self.canvas.zoom, y + i*height/offset - 12*self.canvas.zoom)
            self.coords("argvalue="+j, x - width/2 + 12*self.canvas.zoom, y + i*height/offset)
            i += 2

        if not self.opts["show_kwargs"]: return
        for k in self.kwargs.store:
            self.coords("arg="+k, x - width/2 - node, y - node + i*height/offset, x - width/2 + node, y + node + i*height/offset)
            self.coords("argname="+k, x - width/2 + 12*self.canvas.zoom, y + i*height/offset - 12*self.canvas.zoom)
            self.coords("argvalue="+k, x - width/2 + 12*self.canvas.zoom, y + i*height/offset)
            i += 2


    def move(self, x=0, y=0, wires_only=False):
        if not wires_only:
            for i in self.children():
                self.canvas.move(i, x, y)

        if self.connection or self in self.canvas.output.items:
            self.move_wire()

        for j in self.args["connection"]:
            self.args.store[j].move_wire()

        for k in self.kwargs["connection"]:
            self.kwargs.store[k].move_wire()


    def move_wire(self, *coord):
        if not coord:
            if self in self.canvas.output.items or self == self.canvas.output.log_item:
                coord = (self.canvas.winfo_width(), self.canvas.winfo_height()/2)
            elif self.connection:
                x1, y1, x2, y2 = self.canvas.coords(self.canvas[self.connection[0]].ids["arg=" + self.connection[1]])
                coord = ((x1 + x2)/2, (y1 + y2)/2)
            else:
                self.hide_wire()
                return
        if not self.show_wire:
            self.show_wire = True
            self.config("output", fill=Color.PURPLE)
            self.config("wire", fill=Color.PURPLE, state="normal")

        x,y = coord
        try:
            x1, y1, x2, y2 = self.coords()
        except:
            return
        delta_x = x - (x1 + x2)/2
        node = NODE_SIZE * self.canvas.zoom
        self.coords("wire",
            x2, y1 + (y2 - y1)/2,
            x2 + delta_x /3, (y1 + y2)/2,
            x - delta_x/3, y,
            x - node,y
        )

    def hide_wire(self):
        self.config("wire", state="hidden")
        self.config("output", fill=Color.EMPTY_NODE)
        self.show_wire = False

    def hasargs(self):
        for i in self.args:
            if self.args[i] is None:
                return False
        return True

    def dimensions(self, obj):
        if self.is_callable:
            args, kwargs = self.argspec
            num_args = len(self.argspec[0]) if not self.opts["show_kwargs"] else len(self.argspec[0]) + len(self.argspec[1])
            width = 120 * self.canvas.zoom
            height = max(80, num_args * 6 * 6) * self.canvas.zoom
            return width, height

        char = self.text_dimensions(obj)
        width = min(300, 80 if len(str(obj)) <= 4 else char["width"] * FONT_SIZE) * self.canvas.zoom
        height = (60 + char["height"] * FONT_SIZE) * self.canvas.zoom

        return width, height

    def text_dimensions(self, obj, **kwargs):
        content = self.content(obj, **kwargs)

        if isinstance(obj, dict):
            return {
                "width": min(20, 1 + len(content or "")),
                "height": 1 + content.count("\n")
            }

        return {
            "width": min(20, 1 + len(content or "")),
            "height": 1 + int(len(content or "") / 20)
        }

    def children(self):
        return [self.ids[j] for j in self.ids]

    def config(self, *tags, **kwargs):
        if len(tags) == 1:
            return self.canvas.itemconfig(self.ids[tags[0]], **kwargs)

        for t in tags:
            self.canvas.itemconfig(self.ids[t], **kwargs)

    def coords(self, tag="parent", *args):
        if tag not in self.ids: return False
        return self.canvas.coords(self.ids[tag], *args)

    def position(self):
        x1, y1, x2, y2 = self.coords()
        return (x1 + x2)/2, (y1 + y2)/2

    def destroy(self):
        for i in self.children():
            self.canvas.delete(i)
        del self.canvas[self.name]
        del self



"""
Math Inspector: a visual programming environment for scientific computing
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import sys
from os import path
from ..util.config import BASEPATH
from OpenGL.GL import GL_VERTEX_SHADER, GL_FRAGMENT_SHADER, glGetUniformLocation
from OpenGL.GL import shaders

SHADER_DIR = path.join(BASEPATH, "shader")

class Shader:
	def __init__(self, vert_path, frag_path, uniform=()):
		self.vert_shader = open(path.join(SHADER_DIR, vert_path)).read()
		self.frag_shader = open(path.join(SHADER_DIR, frag_path)).read()
		vert = shaders.compileShader(self.vert_shader, GL_VERTEX_SHADER)
		frag = shaders.compileShader(self.frag_shader, GL_FRAGMENT_SHADER)
		self.program = shaders.compileProgram(vert, frag)

		uniform = uniform if isinstance(uniform, tuple) else (uniform,)
		self.uniform = { i: glGetUniformLocation(self.program, i) for i in uniform }

	def __getitem__(self, key):
		return self.uniform[key]



"""
Math Inspector: a visual programming environment for scientific computing
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import tkinter as tk
from collections import OrderedDict
from ..style import Color
from .button import Button

OPTIONS = {
	"background": Color.BLACK,
	"padx": 12,
	"pady": 8
}

class Popup(tk.Toplevel):
	def __init__(self, app, params, callback, title="", header="", **kwargs):
		OPTIONS.update(kwargs)
		tk.Toplevel.__init__(self, app, **OPTIONS)
		self.app = app
		self.title(title)
		self.callback = callback

		if header:
			tk.Label(self,
				text=header,
				font="Montserrat 16 bold",
				foreground=Color.WHITE,
				background=Color.PURPLE,
				anchor="w",
				padx=8,
				pady=2
			).grid(row=0,column=0, columnspan=2, sticky="nsew", pady=4)

		self.params = OrderedDict()
		for i,val in enumerate(params):
			label = tk.Label(self, text=val["label"], font="Menlo 14", foreground=Color.WHITE, background=Color.BLACK)
			entry = tk.Entry(self, foreground=Color.BLACK, background=Color.VERY_LIGHT_GREY)

			if "value" in val:
				entry.insert("end", str(val["value"]))

			self.params[val["label"]] = entry
			label.grid(row=i+1, column=0, sticky="w", padx=2, pady=4)
			entry.grid(row=i+1, column=1, sticky="w", padx=2, pady=4)

		self.bind("<Return>", self._on_ok)
		self.bind("<Escape>", self._on_cancel)

		btn_area = tk.Frame(self, background=Color.BLACK)
		Button(btn_area, text="Cancel", command=self._on_cancel, width=8).grid(row=0, column=0, padx=8)
		Button(btn_area, text="OK", command=self._on_ok, width=8).grid(row=0, column=1, padx=8)
		btn_area.grid(row=i+2, columnspan=2, pady=12)


	def _on_ok(self, event=None):
		result = {}
		for i in self.params:
			try:
				temp = self.app.console.eval(self.params[i].get())
			except:
				temp = self.params[i].get()
			result[i] = temp
		self.callback(result)
		self.destroy()

	def _on_cancel(self, event=None):
		self.destroy()


"""
Math Inspector: a visual programming environment for scientific computing
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from ..style import Color
from ..util.config import FONT, FONT_SIZE

DOC_TAGS = {
    "name": {
        "font": "Montserrat " + FONT_SIZE["extra-large"] + " bold",
        "foreground": Color.WHITE
    },

    "signature": {
        "font": "Monospace " + FONT_SIZE["small"]
    },

    "function_name": {
        "spacing1": 8,
        "spacing3": 8,
        "font": "Monospace " + FONT_SIZE["medium"]
    },

    "class_name": {
        "font": "Nunito " + FONT_SIZE["extra-large"]
    },

    "section_title": {
        "font": "Montserrat " + FONT_SIZE["large"] + " bold",
        "foreground": Color.YELLOW,
        "spacing1": 0,
        "spacing3": 8
    },

    "module_nav": {
        "font": "Montserrat " + FONT_SIZE["medium"] + " bold",
        "foreground": Color.DARK_GREY,
        "spacing1": 12,
    },

    "module_nav_hover": {
        "foreground": Color.WHITE,
    },

    "root": {
        "font": "Montserrat " + FONT_SIZE["medium"] + " bold",
        "foreground": Color.DARK_ORANGE,
    },

    "root_hover": {
        "foreground": Color.ORANGE,
    },

    "bordered_box": {
        "background": Color.COOL_GREY,
        "lmargin1": 16,
        "rmargin": 16
    },

    "spacing_top": {
        "spacing1": 16
    },

    "spacing_bottom": {
        "spacing3": 16
    },

    "h1": {
        "font": "Montserrat " + FONT_SIZE["extra-large"] + " bold",
        "background": Color.DARK_PURPLE,
        "lmargin1": 16,
        "lmargin2": 8,
        "spacing1": 8,
        "spacing3": 8
    },

    "parameter_name": {
        "spacing1": 8,
        "font": "Monospace "+ FONT_SIZE["default"] + " bold",
        "lmargin1": 24,
        "foreground": Color.ORANGE
    },

    "submodule": {
        "spacing1": 8,
        "spacing3": 8,
        "font": "Monospace "+ FONT_SIZE["default"] + " bold",
        "lmargin1": 24,
        "foreground": Color.ORANGE
    },

    "submodule_hover": {
        "foreground": Color.DARK_ORANGE
    },

    "parameter_type": {
        "font": "Nunito " + FONT_SIZE["extra-large"]
    },

    "parameter_description": {
        "lmargin1": 48,
        "lmargin2": 48,
    },

    "underline_title": {
        "font": "Montserrat " + FONT_SIZE["large"] + " bold",
        "foreground": Color.VERY_LIGHT_GREY,
        "spacing1": 16
    },

    "notes": {
    },

    "references": {
        "font": "Nunito " + FONT_SIZE["small"] + " bold"
    },

    "see_also_title": {
        "font": "Montserrat " + FONT_SIZE["medium"] + " bold",
        "spacing1": 16,
        "spacing3": 8,
    },

    "see_also": {
        "background": Color.HIGHLIGHT_INACTIVE,
        "lmargin1": 16,
        "lmargin2": 32,
    },

    "doc_link": {
        "foreground": Color.BLUE,
    },

    "doc_link_hover": {
        "foreground": Color.LINK_URL,
    },

    "code_sample": {
        "background": Color.CONSOLE_BACKGROUND,
        # "foreground": Color.WHITE,
        "font": FONT,
        "spacing1": 8,
        "spacing2": 0,
        "spacing3": 8,
        "lmargin1": 8,
        "lmargin2": 38
    },

    "code_sample_hover": {
        "background": Color.HIGHLIGHT_INACTIVE
    },

    "code": {
        "foreground": Color.GREY,
        "font": FONT + " bold"
    },

    "console_prompt": {
        "foreground": Color.PROMPT,
        "font": FONT + " bold",
        "selectbackground": "red"
    },

    "link_url": {
        "foreground": Color.BLUE
    },

    "link_url_hover": {
        "foreground": Color.LINK_URL
    },

    "section_heading": {
        "font": "Monospace 20 bold",
        "foreground": Color.BLACK,
        "background": Color.LIGHTER_GREY
    },

    "underline_heading": {
        "font": "Lato " + FONT_SIZE["medium"],
        "foreground": Color.BLACK,
        "underline": "true"
    },

    "bold": {
        "font": "Montserrat " + FONT_SIZE["default"] + " bold",
    },

    "menlo_italic": {
        "font": FONT + " italic",
    },

    "horizontal_rule": {
        "font": "Monospace 1",
        "background": Color.VERY_LIGHT_GREY
    }

}



"""
Math Inspector: a visual programming environment for scientific computing
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import tkinter as tk
import numpy as np
import inspect, platform
from .. import plot
from ..util import fontcolor, instanceof, classname, argspec, numargs, open_editor, vdict, instanceof
from ..util.config import BUTTON_RIGHT, BUTTON_RELEASE_RIGHT, BUTTON_RIGHT_MOTION, HITBOX, ZOOM_IN, ZOOM_OUT, FONTSIZE, FONT_SIZE
from ..style import Color, getimage
from .output import Output
from .item import Item
from ..widget import Popup, Menu, Text

class NodeEditor(vdict, tk.Canvas):
	def __init__(self, app):
		self.app = app
		vdict.__init__(self)
		self.frame = tk.Frame(app, background=Color.BLACK)
		tk.Canvas.__init__(self, self.frame,
			highlightthickness=4,
			highlightbackground=Color.BACKGROUND,
			background=Color.BACKGROUND,
			bd=0)
		self.zoom = 1
		self.pan_position = None
		self.dragitem = None
		self.dragposition = (0,0)
		self.multiselect = {
			"position": None,
			"items": [],
			"box": self.create_rectangle(0,0,0,0, outline=Color.DARK_GREY, dash=(3,3), state="hidden")
		}

		self.pack(side="left", fill="both", expand=True)
		self.output = Output(self)
		self.output.pack(side="right")

		self.selected = None
		self.connect = None
		self.hover = None
		self.hover_input = None
		self.hover_edit = None
		self.edit_item = None
		self.dragstart = (0,0)
		self.has_menu = False
		self.menu = Menu(self.app)

		self.bind("<Button-1>", self._on_button_1)
		self.bind("<B1-Motion>", self._on_multiselect)
		self.bind("<ButtonRelease-1>", self._on_button_release_1)

		self.bind("<MouseWheel>", self._on_mouse_wheel)
		if platform.system() == "Linux":
			self.bind("<Button-4>", lambda event: self._on_mouse_wheel(event, 1))
			self.bind("<Button-5>", lambda event: self._on_mouse_wheel(event, -1))
		self.bind(BUTTON_RELEASE_RIGHT, self._on_button_release_2)
		self.bind(BUTTON_RIGHT_MOTION, self._on_b2_motion)

		self.tag_bind("draggable", "<Button-1>", self._on_drag_start)
		self.tag_bind("draggable", "<ButtonRelease-1>", self._on_drag_stop)
		self.tag_bind("draggable", "<B1-Motion>", self._on_drag)
		self.tag_bind("draggable", BUTTON_RIGHT, self._on_item_menu)
		self.tag_bind("editable", "<Enter>", self._on_enter_editable)
		self.tag_bind("editable", "<Motion>", self._on_motion_editable)
		self.tag_bind("editable", "<ButtonRelease-1>", self._on_click_editable)
		self.tag_bind("editable", "<Leave>", self._on_leave_editable)
		self.tag_bind("editable", "<Button-1>", self._on_drag_start)
		self.tag_bind("editable", "<B1-Motion>", self._on_drag)

		self.tag_bind("input", "<Motion>", self._on_motion_input)
		self.tag_bind("input", "<Enter>", self._on_enter_input)
		self.tag_bind("input", "<Leave>", self._on_leave_input)

		for i in ("draggable", "output", "wire"):
			self.tag_bind(i, "<Enter>", lambda event, tag=i:self._on_item_enter(event, tag))
			self.tag_bind(i, "<Leave>", lambda event, tag=i:self._on_item_leave(event, tag))

		for j in ("output", "wire", "input"):
			self.tag_bind(j, "<Button-1>", lambda event, tag=j:self._on_item_button_1(event, tag))
			self.tag_bind(j, "<B1-Motion>", lambda event, tag=j:self._on_item_b1_motion(event, tag))
			self.tag_bind(j, "<ButtonRelease-1>", lambda event, tag=j:self._on_item_button_release_1(event, tag))

		self.output.node.bind("<Button-1>", self._on_output_button_1)
		self.output.node.bind("<B1-Motion>", self._on_output_b1_motion)
		self.output.node.bind("<ButtonRelease-1>", self._on_output_button_release_1)


	def setitem(self, name, update_value=False, coord=None, is_output_item=False):
		if name not in self:
			self[name] = Item(self, name, coord=coord or self.get_pointer(random=True))
			self.scale_font()
			if len(self.store) == 1:
				self.app.menu.setview("node_editor", True)
			return

		item = self[name]
		if update_value:
			item.value(self.app.objects[name])
			return

		coord = item.position()
		connection = item.connection
		self[name].destroy()
		if not coord:
			coord = self.get_pointer(random=True)
		self[name] = Item(self, name,
			coord=coord,
			args=item.args,
			kwargs=item.kwargs,
			connection=connection,
			opts=item.opts,
			is_output_item=is_output_item
		)
		self.scale_font()

	def select(self, name):
		if name is None:
			self.selected = None
			return
		if self.selected == name: return
		if self.selected:
			self[self.selected].config("parent", fill=Color.BLACK)
		self[name].config("parent", fill=Color.VERY_LIGHT_PURPLE)
		self.selected = name

	def _on_item_enter(self, event, tag):
		if self.is_busy(): return
		closest = self.get_closest(event.x, event.y, tag) or self.find_closest(event.x, event.y)
		item = self[self.getname(closest)]
		self.hover = closest, item

		if tag == "draggable":
			if item.name != self.selected:
				item.config("parent", fill=Color.LIGHT_PURPLE)
		elif tag in ("output", "wire"):
			item.config("output", fill=Color.HOVER if item.connection else Color.LIGHT_BLACK)
			if item.connection:
				name, argname = item.connection
				self[name].config("arg="+argname, fill=Color.HOVER)
			elif item in self.output.items:
				self.output.hover()
			item.config("wire", "output", fill=Color.HOVER)

	def _on_item_leave(self, event, tag):
		if self.is_busy() or not self.hover: return
		closest, item = self.hover
		self.hover = None

		if tag == "draggable":
			item.config("parent", fill=Color.BLACK if self.getname(closest) != self.selected else Color.VERY_LIGHT_PURPLE)
		elif tag in ("output", "wire"):
			if item.connection:
				name, argname = item.connection
				self[name].config("arg="+argname, fill=Color.PURPLE)
				item.config("output", fill=Color.PURPLE)
				item.config("wire", fill=Color.PURPLE)
			elif item in self.output.items:
				self.output.hover(False)
				item.config("output", fill=Color.PURPLE)
				item.config("wire", fill=Color.PURPLE)
			else:
				item.config("output", fill=Color.EMPTY_NODE)

	def _on_mouse_wheel(self, event, delta=None):
		event_delta = delta or event.delta
		x, y = self.get_pointer()
		if event_delta > 0:
			self.scale("all", x,y, ZOOM_IN, ZOOM_IN)
			self.zoom *= ZOOM_IN
		else:
			self.scale("all", x,y, ZOOM_OUT, ZOOM_OUT)
			self.zoom *= ZOOM_OUT
		self.scale_font()
		self.scale_width()

		if self.output.items:
			for i in self.output.items:
				i.move_wire()

		if self.edit_item:
			self.edit_item.entry.zoom()

	def _on_motion_input(self, event):
		closest = self.find_closest(event.x, event.y)[0]
		if self.hover_input or "input" not in self.gettags(closest): return
		self._on_enter_input(event)

	def _on_enter_input(self, event):
		closest = self.find_closest(event.x, event.y)[0]
		input_item = self[self.getname(closest)]
		argname = self.getname(closest, "arg")
		if argname in input_item.args["connection"]:
			item = input_item.args["connection"][argname]
		elif argname in input_item.kwargs["connection"]:
			item = input_item.kwargs["connection"][argname]
		else:
			return
		self.itemconfig(closest, fill=Color.HOVER)
		item.config("wire", "output", fill=Color.HOVER)
		self.hover_input = closest, item

	def _on_leave_input(self, event):
		if not self.hover_input: return

		closest, item = self.hover_input
		self.itemconfig(closest, fill=Color.PURPLE)
		item.config("wire", "output", fill=Color.PURPLE)
		self.hover_input = None

	def _on_b2_motion(self, event):
		if self.pan_position is None:
			self.pan_position = (event.x, event.y)

		x, y = event.x - self.pan_position[0], event.y - self.pan_position[1]
		self.move("all", x, y)
		self.pan_position = (event.x, event.y)

		if self.output.items:
			for i in self.output.items:
				i.move_wire()

	def _on_item_button_1(self, event, tag):
		closest = self.find_closest(event.x, event.y)[0]
		item = self[self.getname(closest)]

		if tag == "input":
			argname = self.getname(closest, "arg")
			if argname in item.args["connection"]:
				item = item.args["connection"][argname]
			elif argname in item.kwargs["connection"]:
				item = item.kwargs["connection"][argname]
			else:
				return

		if item.connection:
			name, argname = item.connection
			self[name].config("arg="+argname, fill=Color.EMPTY_NODE)

		item.config("wire", "output", fill=Color.PURPLE)
		item.move_wire(event.x, event.y)
		self.connect = (item, None, None, None)

	def _on_output_button_1(self, event):
		if not self.output.items: return
		self.connect = (self.output.items[0], None, None, True)
		self.output.disconnect(self.output.items[0])

	def _on_output_b1_motion(self, event):
		if not self.connect: return
		item, output, argname, was_in_output = self.connect
		item.move_wire(
			self.winfo_pointerx() - self.winfo_rootx(),
			self.winfo_pointery() - self.winfo_rooty()
		)

	def _on_output_button_release_1(self, event):
		if not self.connect: return
		item, output, argname, was_in_output = self.connect
		item.hide_wire()

	def _on_item_b1_motion(self, event, tag):
		if not self.connect: return

		item, output, argname, was_in_output = self.connect

		input_id = self.get_closest(event.x, event.y, "input")
		in_output = self.winfo_containing(event.x_root, event.y_root) in (self.frame, self.output.node)
		in_items = (self.winfo_containing(event.x_root, event.y_root) == self)

		if was_in_output and not in_items:
			self.output.hover(True)
			item.move_wire(event.x, event.y)
			return
		elif item in self.output.items and in_items:
			self.output.disconnect(item)

		if input_id:
			x1,y1,x2,y2 = self.coords(input_id[0])
			item.move_wire((x1 + x2)/2, (y1 + y2)/2)
			output = self[self.getname(input_id)]
			argname = self.getname(input_id, "arg")
			output.config("arg="+argname, fill=Color.HOVER)
			self.connect = item, output, argname, in_output
			return
		else:
			if item.connection:
				item.disconnect()

			item.move_wire(event.x, event.y)
			if output and argname:
				if argname in output.args["connection"] or argname in output.kwargs["connection"]:
					output.config("arg="+argname, fill=Color.PURPLE)
				else:
					output.config("arg="+argname, fill=Color.EMPTY_NODE)
				output = argname = None

		self.connect = item, output, argname, in_output

	def _on_item_button_release_1(self, event, tag):
		if self.connect:
			item, output, argname, in_output = self.connect

			if in_output:
				self.output.connect(item)

			if not output:
				item.disconnect()
			elif argname in output.args:
				output.args[argname] = item
			elif argname in output.kwargs:
				output.kwargs[argname] = item

			self.connect = None

	def is_busy(self):
		return bool(self.dragitem
			or self.connect
			or self.multiselect["items"])

	def _on_delete(self, key=None):
		if self.multiselect["items"]:
			for i in self.multiselect["items"]:
				del self.app.objects[i]
			self.multiselect["items"].clear()
		elif key:
			del self.app.objects[key]

	def _on_multiselect(self, event):
		if self.connect: return

		for name in self.multiselect["items"]:
			self[name].config("parent", fill=Color.BLACK)

		if self.dragitem or not self.multiselect["position"]: return

		x,y = self.multiselect["position"]
		overlapping = self.find_overlapping(x,y,event.x,event.y)
		self.itemconfig(self.multiselect["box"], state="normal")
		self.coords(self.multiselect["box"], x, y, event.x, event.y)

		self.multiselect["items"].clear()

		for i in overlapping:
			name = self.getname(i)
			if name and name not in self.multiselect["items"]:
				self.multiselect["items"].append(name)

		for name in self.multiselect["items"]:
			self[name].config("parent", fill=Color.VERY_LIGHT_PURPLE)


	def _on_motion_editable(self, event):
		closest = self.get_closest(event.x, event.y, "editable")
		if self.hover_edit or "editable" not in self.gettags(closest): return
		self._on_enter_editable(event)

	def _on_enter_editable(self, event, closest=None):
		closest = closest or self.find_closest(event.x, event.y)[0]
		if "editable" not in self.gettags(closest): return

		item = self[self.getname(closest)]
		color = self.itemconfig(closest)["fill"][4]
		self.itemconfig(closest, fill=Color.WHITE)
		self.hover_edit = closest, item, color

	def _on_leave_editable(self, event):
		if self.hover_edit:
			closest, item, color = self.hover_edit
			self.itemconfig(closest, fill=color)
			self.hover_edit = None

	def _on_click_editable(self, event):
		if self.hover_edit and self.dragstart == (event.x, event.y):
			closest, item, color = self.hover_edit
			self.edit_item = item
			item.entry.edit(closest)

	def _on_drag_start(self, event):
		self.dragitem = self.find_closest(event.x, event.y)[0]
		self.dragposition = (event.x, event.y)
		items = self.find_siblings(self.dragitem) #REFACTOR - use item.ids
		for i in items:
			self.tag_raise(i)

		self.dragstart = (event.x, event.y)
		if "editable" in self.gettags(self.dragitem):
			self._on_enter_editable(event, self.dragstart)

	def _on_drag(self, event):
		name = self.getname(self.dragitem)
		if name not in self: return

		item = self[name]
		if "editable" in self.gettags(self.dragitem):
			delta_x = event.x - self.dragposition[0]
			delta_y = event.y - self.dragposition[1]
			self.dragposition = (event.x, event.y)
			key = self.getname(self.dragitem, "argvalue")
			if key == "<value>":
				if isinstance(item.obj, int):
					self.app.objects[item.name] += int(np.sign(delta_x) - np.sign(delta_y))
				elif isinstance(item.obj, float):
					num_decimals = 1 + len(str(item).split(".")[1])
					self.app.objects[item.name] = round(self.app.objects[item.name] + (delta_x - delta_y) / 100, num_decimals)
			else:
				args = item.args if key in item.args else item.kwargs
				if isinstance(args[key], int):
					args[key] = args[key] + int(np.sign(delta_x) - np.sign(delta_y))
				elif isinstance(args[key], float):
					num_decimals = 1 + len(str(item).split(".")[1])
					args[key] = round(args[key] + (delta_x - delta_y) / 100, num_decimals)
			if "<value>" in item.args["connection"]:
				item.args["connection"]["<value>"].disconnect()
			item.resize()
			return

		x, y = self.dragposition
		if self.multiselect["items"]:
			for i in self.multiselect["items"]:
				self[i].move(event.x - x, event.y - y)
		else:
			item.move(event.x - x, event.y - y)

		self.dragposition = (event.x, event.y)


	def _on_drag_stop(self, event):
		if self.dragstart == self.dragposition:
			self.select(self.getname(self.dragitem))
		self.dragitem = None
		self.dragposition = (0,0)

	def _on_button_1(self, event):
		if self.edit_item:
			self.edit_item.entry.finish()

		if self.dragitem or self.connect: return
		if self.hover_edit:
			self[self.getname(self.hover_edit)].entry.finish()
			self.hover_edit = None
			return
		for name in self.multiselect["items"]:
			self[name].config("parent", fill=Color.BLACK)
		self.select(None)

		# if self.dragitem: return

		self.multiselect["items"].clear()
		self.multiselect["position"] = (event.x, event.y)
		self.tag_raise(self.multiselect["box"])

	def _on_button_release_1(self, event):
		self.multiselect["position"] = None
		self.itemconfig(self.multiselect["box"], state="hidden")

	def _on_button_release_2(self, event):
		if not self.pan_position and not self.has_menu:
			self.menu.show(event, [{
				"label": "Add Object",
				"font": "Nunito " + FONT_SIZE["extra-small"] + " bold",
				"foreground": Color.DARK_ORANGE,
				"state": "disabled"
			}] + self.app.menu.object_menu)
		self.has_menu = False
		self.pan_position = None

	# REFACTOR: move the bulk of this into its own class with a __call__ attr so its super simple to call in here
	def _on_item_menu(self, event):
		self.has_menu = True
		item = self[self.getname(self.find_closest(event.x, event.y))]
		if self.multiselect["items"]:
			self.menu.show(event, [{
				"label": "Delete selected items...",
				"command": self._on_delete
			}])
			return

		methods = []
		graph = []
		extras = []
		obj = self.app.objects[item.name]
		value = item.value()
		item_methods = [i for i in dir(obj) if i[:1] != "_" and callable(getattr(obj, i)) and isinstance(i, str)]
		if item_methods:
			methods.append({ "separator": None })
			for fn in item_methods:
				attr = getattr(item.obj, fn)
				methods.append({
					"label": fn,
					"command": lambda fn=fn, attr=attr, item=item: self.run_method(fn, attr, item)
				})

		window = plot.get_window(value)
		if item in self.output.items:
			extras.append({
				"label": "Disconnect Output",
				"command": lambda: self.output.disconnect(item)
			})
		elif not window:
			extras.append({
				"label": "Show Output",
				"command": lambda: self.output.connect(item)
			})
		else:
			extras.append({
				"label": "Plot",
				"command": lambda: self.output.connect(item),
				"state": "normal" if (plot.active_window == window or not plot.is_active()) else "disabled"
			})

		extras.append({
			"separator": None
		})

		if self.app.animate.can_animate(item):
			extras.append({
				"label": "Animate",
				"command": lambda: self.app.animate(item.name)
			})

		if len(item.kwargs) > 0 and len(item.kwargs["connection"]) == 0:
			extras.append({
				"label": "Hide kwargs" if item.opts["show_kwargs"] else "Show kwargs",
				"command": lambda: item.option("show_kwargs")
			})

		if help.getobj(item.obj) is not None:
			extras.append({
				"label": "View Doc",
				"command": lambda: help(item.obj)
			})

		try:
			file = inspect.getsourcefile(item.obj)
		except:
			file = None

		if file:
			extras.append({
				"label": "View Source Code",
				"command": lambda: open_editor(file)
			})

		# graph.append({
		# 	"label": "Set color     " + item.opts["line_color"] if item.opts["line_color"] != Color.BLUE else "Set color",
		# 	"command": lambda: Popup(self.app, item.name + ".set_color", obj=lambda color: item.option("line_color", color), canvas_item=item)
		# })

		self.menu.show(event, extras + methods + [{
			"separator": None
		},{
			"label": "Delete " + item.name,
			"command": lambda: self._on_delete(item.name)
		}])

	def run_method(self, name, attr, item):
		args, kwargs = argspec(attr)
		args = [i for i in args if i not in ("self")]
		if len(args) + len(kwargs) > 0:
			Popup(self.app, [{
				"label": i
			} for i in args] + [{
				"label": k,
				"value": kwargs[k]
			} for k in kwargs],
			lambda params: self._method_result(name, attr, item, params),
			title="Run Item Method",
			header=name)
		else:
			self._method_result(name, attr, item)

	def _method_result(self, name, attr, item, params={}):
		if params:
			try:
				argcount = numargs(attr)
			except:
				argcount = len(params) - 1
			keys = list(params.keys())[1 + argcount:]
			args = [params[i] for i in params][:argcount]
			kwargs = { i:params[i] for i in keys }
			if kwargs:
				result = attr(*args, **kwargs)
			else:
				result = attr(*[i for i in list(params.values()) if i != ""])
		else:
			result = attr()

		item.value(self.app.objects[item.name])
		if instanceof(result, (int, float, complex, bool)):
			self.output.show(result)
		elif result:
			self.app.objects.setobj(name, result, create_new=True)

	def get_pointer(self, random=False):
		if random:
			x = np.random.rand() * self.winfo_width()
			y = np.random.rand() * self.winfo_height()
		else:
			x = self.canvasx(self.winfo_pointerx() - self.winfo_rootx())
			y = self.canvasy(self.winfo_pointery() - self.winfo_rooty())
		return x,y

	def scale_font(self):
		font_items = []
		for i in self:
			font_items.extend([self[i].ids["name"], self[i].ids["class"]])
			if "value" in self[i].ids:
				 font_items.append(self[i].ids["value"])
			for j in self[i].ids:
				if j[:7] == "argname" or j[:9] == "kwargname" or j[:8] == "argvalue" or j[:10] == "kwargvalue":
					font_items.append(self[i].ids[j])

		for i in font_items:
			font = self.itemconfig(i)["font"][3].split(" ")
			tags = self.gettags(i)
			fontsize = FONTSIZE
			for t in tags:
				if t[:9] == "fontsize=":
					if "stop_scaling" in tags and self.zoom < 1:
						fontsize = t[9:]
					else:
						fontsize = str(int(int(t[9:]) * self.zoom) or 1)
			self.itemconfig(i, font=font[0] + " " + fontsize + ("" if len(font) < 3 else " " + font[2]))

	def scale_width(self):
		width_items = [self[i].ids["wire"] for i in self]
		for i in width_items:
			self.itemconfig(i, width=1 + int(self.zoom))

	def find_siblings(self, item):
		name = self.getname(item)
		return self.find_withtag("name=" + name) if name else [item]

	def getname(self, item, key="name"):
		for t in self.gettags(item):
			if t[:len(key) + 1] == key + "=":
				return t[len(key) + 1:]
		return None

	def find_withtag(self, *args):
		if len(args) <= 1:
			return super(NodeEditor, self).find_withtag(args[0])

		result = self.find_withtag(args[0])
		for i in range(1, len(args)):
			result = [j for j in self.find_withtag(args[i]) if j in result]

		if len(result) == 1:
			return result[0]

		return result

	def get_closest(self, x,y, tag=None):
		overlapping = self.find_overlapping(
			x - HITBOX * self.zoom, y - HITBOX * self.zoom,
			x + HITBOX * self.zoom, y + HITBOX * self.zoom
		)

		if tag == None:
			return overlapping

		return tuple([i for i in self.find_closest(x,y) if i in self.find_withtag(tag)])



"""
Math Inspector: a visual programming environment for scientific computing
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import numpy as np
import scipy, math, builtins, os, sys
from functools import partial
from tkinter import filedialog

from . import plot, doc, examples
from .util import binop
from .widget import Popup, Menu
from .style import Color
from .util import BUILTIN_FUNCTION, BUILTIN_CLASS, BUILTIN_CONSTANT
from .util.common import open_editor
from .util.config import CONTROL_KEY, BASEPATH, BUILTIN_FUNCTION, BUILTIN_CLASS, BUILTIN_CONSTANT
from .console import builtin_print

TRIG_FUNCTIONS = [i for i in ("acos", "acosh", "asin", "asinh", "atan", "atan2", "atanh", "cos", "cosh", "degrees", "sin", "sinh", "tan", "tanh")]
MATH_FUNCTIONS = [i for i in dir(math) if callable(getattr(math, i)) and i not in TRIG_FUNCTIONS + ["pow"] and i[:1] != "_"]
MATH_FUNCTIONS.append("power")
MATH_FUNCTIONS.sort()
MATH_CONSTANTS = [i for i in dir(math) if i not in MATH_FUNCTIONS + TRIG_FUNCTIONS and i[:1] != "_"]
SASHPOS_MIN = 20

class MainMenu(Menu):
	def __init__(self, app):
		self.project_menu = [{
			"label": "New File",
			"command": self.new_file
		},{
			"label": "Add File...",
			"command": app.modules.addfile
		},{
			"label": "Add Folder...",
			"command": app.modules.addfolder
		},{
			"label": "Import Module...",
			"command": lambda event=None: help.import_module(self.import_module)
		}]

		self.object_menu = [{
			"label": "Operator",
			"menu": [{
				"label": i,
				"command": lambda key=i: self.create_object(key, binop)
			} for i in dir(binop) if i[0] != "_"]
		},{
			"label": "Math",
			"menu": [{
				"label": "Constant",
				"menu":[{
					"label": i,
					"command": lambda key=i: self.create_object(key, np if hasattr(np, key) else math)
				} for i in MATH_CONSTANTS]
			},{
				"separator": None
			},{
				"label": "Trig",
				"menu":[{
					"label": i,
					"command": lambda key=i: self.create_object(key, np if hasattr(np, key) else math)
				} for i in TRIG_FUNCTIONS]
			},{
				"separator": None
			}] + [{
				"label": i,
				"command": lambda key=i: self.create_object(key, np if hasattr(np, key) else math)
			} for i in MATH_FUNCTIONS]
		},{
			"label": "Builtin Class",
			"menu": [{
				"label": str(i),
				"command": lambda key=i: self.create_object(key, builtins)
			} for i in BUILTIN_CLASS if i[:1] != "_" and not i[:1].isupper() and "Error" not in i and "Warning" not in i]
		},{
			"label": "Builtin Function",
			"menu": [{
				"label": str(i),
				"command": lambda key=i: self.create_object(key, module=builtins)
			} for i in BUILTIN_FUNCTION if i[:1] != "_" and not i[:1].isupper()]
		},{
			"label": "Examples",
			"menu": [{
				"label": str(i).replace("_", " ").capitalize(),
				"command": lambda key=i: self.create_object(key, module=examples)
			} for i in dir(examples) if i[:1] != "_" and not i[:1].isupper() and i != "np"]
		}]

		Menu.__init__(self, app, [{
			"label": "File",
			"menu": [{
					"label": "New                    ",
					"command": app.project.new,
					"accelerator": CONTROL_KEY + "+n",
				},
				{
					"label": "Open...            ",
					"command": lambda event=None: app.project.load(None),
					"accelerator": CONTROL_KEY + "+o"
				},
				{
					"label": "Save            ",
					"command": lambda event=None: app.project.save(app.modules.rootfolder),
					"accelerator": CONTROL_KEY + "+s"
				},
				{
					"label": "Save As...            ",
					"command": lambda event=None: app.project.save(None),
					"accelerator": CONTROL_KEY + "+Shift+s"
				}]
			},{
				"label": "Project",
				"menu": self.project_menu
			},{
				"label": "Object",
				"menu": self.object_menu
			},{
				"label": "Plot",
				"menu": [{
					"label": "Show Grid Lines                      ✓",
					"command": partial(self.plotconfig, "show_grid"),
				},{
					"label": "Show Range                             ",
					"command": lambda: self.plotconfig("show_range")
				}],
			},
			{
				"label": "View",
				"menu": [{
					"label": "Show Sidebar",
					"command": lambda: self.setview("sidebar")
				},{
					"label": "Show Node Editor",
					"command": lambda: self.setview("node_editor")
				},{
					"label": "Hide Console",
					"command": lambda: self.setview("console")
				}]
			},
			{
				"label": "Help",
				"menu": [{
					"label": "Getting Started",
					"command": lambda: help(doc.manual.GettingStarted)
				},{
					"label": "User Manual",
					"command": lambda: help(doc.manual)
				},{
					"label": "Examples",
					"command": lambda: help(examples)
				},
				# {
				##### for some reason this is failing in pyinstaller, can include this with .rst files maybe
				# 	"label": "Numpy Doc",
				# 	"command": lambda: help(np.doc)
				# },
				{
					"separator": None
				},{
					"label": "Browse Modules...",
					"command": lambda: help.browse()
				}
				# ,{
				# 	"label": "Scipy Tutorial",
				# 	"command": lambda: help(scipy)
				# }
				# ,{
				# 	"label": "Builtin Functions",
				# 	"command": lambda: help(builtins)
				# }
				]
			}])

		self.app = app
		self.is_sidebar_visible = False
		self.has_hidden_panel = False

		if hasattr(sys, "_MEIPASS"):
			app.createcommand('tkAboutDialog', lambda: help(os.path.join(BASEPATH, "ABOUT.md"), "About Math Inspector"))

		app.side_view.bind("<Configure>", self.on_config_sidebar)
		app.node.output.bind("<Configure>", self.on_config_vertical_panel)

	def on_config_sidebar(self, event):
		sashpos = self.app.horizontal_panel.sashpos(0)
		if self.is_sidebar_visible and sashpos > SASHPOS_MIN: return

		if sashpos <= SASHPOS_MIN:
			self._["View"].entryconfig(0, label="Show Sidebar")
			self.is_sidebar_visible = False
		else:
			self._["View"].entryconfig(0, label="Hide Sidebar")
			self.is_sidebar_visible = True

	def on_config_vertical_panel(self, event):
		sashpos = self.app.vertical_panel.sashpos(0)
		height = self.app.winfo_height()
		if SASHPOS_MIN < sashpos < height - SASHPOS_MIN:
			if self.has_hidden_panel:
				self.has_hidden_panel = False
				self._["View"].entryconfig(1, label="Hide Node Editor")
				self._["View"].entryconfig(2, label="Hide Console")
			return

		self.has_hidden_panel = True
		if sashpos <= SASHPOS_MIN:
			self._["View"].entryconfig(1, label="Show Node Editor")
			self._["View"].entryconfig(2, label="Hide Console")
		else:
			self._["View"].entryconfig(1, label="Hide Node Editor")
			self._["View"].entryconfig(2, label="Show Console")

	def create_object(self, name, module):
		self.app.objects.setobj(name, getattr(module, name), create_new=True)

	def setview(self, key, force_open=False):
		h_sashpos = self.app.horizontal_panel.sashpos(0)
		v_sashpos = self.app.vertical_panel.sashpos(0)

		if key == "sidebar":
			self.is_sidebar_visible = not self.is_sidebar_visible
			if self.is_sidebar_visible:
				self._["View"].entryconfig(0, label="Hide Sidebar")
				self.app.horizontal_panel.sashpos(0,240)
			elif not force_open:
				self._["View"].entryconfig(0, label="Show Sidebar")
				self.app.horizontal_panel.sashpos(0,0)
		elif key == "node_editor":
			if v_sashpos <= SASHPOS_MIN:
				self.app.vertical_panel.sashpos(0, int(self.app.winfo_height()/2))
				self._["View"].entryconfig(1, label="Hide Node Editor")
			elif not force_open:
				self.app.vertical_panel.sashpos(0, 0)
				self._["View"].entryconfig(1, label="Show Node Editor")
		elif key == "console":
			height = self.app.winfo_height()
			if v_sashpos >= height - SASHPOS_MIN:
				self.app.vertical_panel.sashpos(0, int(height/2))
			elif not force_open:
				self.app.vertical_panel.sashpos(0, int(height))
		elif key == "modules":
			self.app.side_view.select(0)
			if h_sashpos <= 10:
				self.app.horizontal_panel.sashpos(0,240)
		elif key == "objects":
			self.app.side_view.select(1)
			if h_sashpos <= 10:
				self.app.horizontal_panel.sashpos(0,240)
			# else:
			# 	self.app.vertical_panel.sashpos(0, 0)

	def new_file(self):
		file = filedialog.asksaveasfilename(defaultextension=".py")
		if not file: return
		f = open(file, "a")
		f.write("")
		f.close()
		self.app.modules.addfile(file)
		open_editor(file)

	def import_module(self, module, alias="", open_folders=False):
		self.app.modules[alias or module] = __import__(module)
		if open_folders:
			self.app.focus()
			key = alias or module
			self.app.modules.selection_set(key)
			self.app.modules.item(key, open=True)

	def plotconfig(self, key=None, **kwargs):
		opts = plot.config()
		if not opts: return # BUG: make sure this works when loading a project

		if key is None:
			plot.config(**kwargs)
			if "show_grid" in kwargs:
				self._["Plot"].entryconfig(0, label="Show Grid Lines                      " + ("✓" if opts["show_grid"] else ""))
			elif "show_range" in kwargs:
				self._["Plot"].entryconfig(1, label="Show Range                           " + ("✓" if opts["show_range"] else ""))
			return

		plot.config(**{ key : not opts[key] })
		if key == "show_grid":
			self._["Plot"].entryconfig(0, label="Show Grid Lines                      " + ("✓" if opts[key] else ""))
		elif key == "show_range":
			self._["Plot"].entryconfig(1, label="Show Range                           " + ("✓" if opts[key] else ""))



	def set_plot_resolution(self, val):
		self.plotconfig(resolution=val)



"""
The math inspector console extends the traditional python interpreter, and has many quality of
life improvements such as syntax highlighting and a wide variety of hotkeys.

There are only two built in functions in math inspector which are not available in a normal python
environment: `app` and `plot`.  The `app` object opens up the entire math inspector codebase,
allowing you to interact with the any part of the app directly from the console.  This can be
very useful when prototyping new features.

For example, to toggle the left hand sidebar open and closed, you can use the command:

>>> app.menu.setview("sidebar")

Hotkeys
---
`up-arrow`: toggle history up

`down-arrow`: toggle history down

`Tab`: autocomplete

`Cmd+d`: highlight the currently selected word

"""
from .interpreter import Interpreter
from .builtin_print import builtin_print


"""
Math Inspector: a visual programming environment for scientific computing
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import keyword, platform
from .history import History
from .builtin_print import builtin_print
from .autocomplete import AutoComplete
from ..util.config import FONT, PROMPT_FONTSIZE as FONTSIZE
from ..widget import Text
from ..style import Color

class Prompt(Text):
	def __init__(self, console, frame):
		Text.__init__(self, frame,
			background=Color.DARK_BLACK,
			font=FONT,
			padx=0,
			pady=0,
			height=2)

		self.console = console
		self.is_on_bottom = False

		self.autocomplete = AutoComplete(self)
		self.history = History(self)
		self.yscroll = 1.0

		self.bind("<Key>", self._on_key)
		self.bind("<<Modify>>", self.on_modify)
		self.bind("<<MarkSet>>", self.on_mark_set)
		self.bind("<<Selection>>", self.on_text_selection)
		self.bind("<FocusIn>", self.on_focus_in)
		self.console.bind("<<Scroll>>", self._on_scroll_log)
		self.lift()

	def __call__(self):
		self.delete("1.0", "end")
		self.insert("end", "...  " if self.console.buffer else ">>>  ", "console_prompt")
		self.console.see("end")
		self.move()
		self.focus()
		self.edit_reset() # NOTE: this resets the tkinter undo stack

	def get(self, start="1.5", end="end-1c"):
		return super(Prompt, self).get(start, end)

	def _on_key(self, event):
		self.console.remove_selection()
		if event.keysym == "Return":
			self.push(self.get())
			return "break"

		if event.keysym == "Tab" and self.get().strip():
			self.autocomplete()
			return "break"

		if event.keysym in ("Up", "Down"):
			self.history.toggle(-1 if event.keysym == "Up" else +1)
			return "break"

		line, col = self.index("insert").split(".")
		if event.keysym in ("BackSpace", "Left"):
			selected_range = self.tag_ranges("sel")
			if selected_range and int(col) == 5 and event.keysym == "BackSpace":
				self.delete(*selected_range)
				return "break"
			elif int(col) < 6:
				builtin_print("\a")
				return "break"

		tag_ranges = self.tag_ranges("sel")
		if tag_ranges and event.keysym in ("parenleft", "quotedbl", "quoteright", "bracketleft"):
			content = self.get(*tag_ranges)
			self.delete(*tag_ranges)

			if event.keysym == "parenleft":
				self.insert(tag_ranges[0], "(" + content + ")")
			elif event.keysym == "quotedbl":
				self.insert(tag_ranges[0], "\"" + content + "\"")
			elif event.keysym == "quoteright":
				self.insert(tag_ranges[0], "'" + content + "'")
			elif event.keysym == "bracketleft":
				self.insert(tag_ranges[0], "[" + content + "]")
			return "break"

		ctrl = (event.state & 0x4) != 0
		meta = (event.state & 0x8) != 0
		is_mod = (ctrl or meta) and platform.system() != "Windows"
		if is_mod:
			if event.char == "v":
				return self._on_paste()

			if event.char == "d":
				self.tag_add("sel", "insert wordstart", "insert wordend")
				return "break"

		## TODO - reenable this, designed to emulate the ipython ? functionality
		# if event.keysym == "question":
		# 	try:
		# 		obj = self.console.eval(self.get())
		# 	except:
		# 		builtin_print("\a")
		# 		return "break"

		# 	if help.getobj(obj):
		# 		help(obj)
		# 	else:
		# 		builtin_print("\a")
		# 	return "break"

	def push(self, s):
		if s or not self.console.buffer:
			if self.console.get("1.0", "end").strip() and not self.console.buffer:
				self.console.insert("end", "\n")
			self.console.insert("end", "...  " if self.console.buffer else ">>>  ", "console_prompt")
			self.console.insert("end", s + "\n", syntax_highlight=True)
		self.console.push(s)

	def on_configure_log(self, event):
		self.console.config(height=self.console.frame.winfo_height() - 2*FONTSIZE)
		if self.is_on_bottom and not self.on_bottom():
			self.is_on_bottom = False

		if self.winfo_ismapped() and self.console.yview() != 1.0:
			self.is_on_bottom = False
		self.move()

	def _on_scroll_log(self, event):
		if not self.is_on_bottom: return

		yscroll = self.console.yview()[1]
		if yscroll < 1:
			self.pack_forget()
		elif not self.winfo_ismapped():
			self.is_on_bottom = False
			self.move()
		self.yscroll = yscroll

	def move(self, event=None):
		if self.is_on_bottom:
			return

		if self.on_bottom():
			self.is_on_bottom = True
			self.pack(side="bottom", fill="x", expand=True, before=self.console)
			self.config(height=2)
			# self.console.pack_forget()
			self.console.pack(fill="both", expand=True)
			self.console.config(height=self.console.winfo_height()/FONTSIZE)
			self.console.see("end")
		else:
			self.place(x=0, y=FONTSIZE*self.linecount(), width=self.console.winfo_width())

	def on_bottom(self):
		count = self.linecount()
		if self.is_on_bottom:
			return FONTSIZE*(1+count) > self.console.frame.winfo_height()
		return FONTSIZE*(1+count) > self.console.winfo_height()

	def linecount(self):
		count = self.console.count("1.0", "end", "displaylines")[0]
		if count == 1:
			return 0
		return count

	def on_modify(self, event):
		self.syntax_highlight("end-1c linestart+5c", "end-1c")

	def on_mark_set(self, event):
		line, col = self.index("insert").split(".")

		if int(col) < 5:
			self.mark_set("insert", line + ".5")

	def on_text_selection(self, event):
		selected = self.tag_ranges("sel")
		if len(selected) == 0: return

		line, col = str(selected[0]).split(".")
		endline = self.index("end-1c").split(".")[0]
		if line == endline and int(col) < 5:
			self.tag_remove("sel", selected[0], str(line) + ".5")

	def _on_paste(self, content=None):
		content = content or self.clipboard_get()
		if not content or "\n" not in content: return

		tag_ranges = self.tag_ranges("sel")
		if tag_ranges:
			self.delete(*tag_ranges)

		self.console.write(">>>  " + content, syntax_highlight=True)
		try:
			self.console.exec(content)
		except:
			self.console.showtraceback()
		return "break"

	def on_focus_in(self, event):
		self.console.remove_selection()

	def _on_button_right(self, event):
		tag_ranges = self.tag_ranges("sel")
		items = []
		if tag_ranges:
			items.extend([{
				"label": "Copy",
				"command": lambda: self.clipboard_append(self.get(*tag_ranges))
			}, {
				"label": "Paste",
				"command": self._on_paste
			}])
		else:
			items.append({
				"label": "Paste",
				"command": self._on_paste
			})

		items.extend([{
			"label": "Clear Log",
			"command": clear
		}, {
			"label": "Keyword List",
			"menu": [{
				"label": str(i),
				"command": lambda key=i: self.insert("insert", key)
			} for i in keyword.kwlist]
		}])

		self.menu.show(event, items)



"""
class that manages the command history feature, which can be accessed from the console by
using the up and down arrow keys.
"""
"""
Math Inspector: a visual programming environment for scientific computing
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from .builtin_print import builtin_print

class History:
	def __init__(self, prompt):
		self.cmds = []
		self.prompt = prompt
		self.i = 0

	def toggle(self, num):
		self.i += num
		if self.i < 0:
			self.i = 0
			builtin_print("\a")
			return

		self.prompt.delete("1.5", "end")
		if self.i > len(self.cmds):
			self.i = len(self.cmds)
			builtin_print("\a")
		elif self.i < len(self.cmds):
			self.prompt.insert("1.5", self.cmds[self.i])

	def clear(self):
		self.cmds.clear()
		self.i = 0

	def append(self, obj):
		self.cmds.append(obj)
		self.i = len(self.cmds)

	def extend(self, *args):
		self.cmds.extend([i.rstrip() for i in args])
		self.i = len(self.cmds)

	def __repr__(self):
		return str(self.cmds)


"""
Math Inspector: a visual programming environment for scientific computing
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import tkinter as tk
from tkinter import ttk
import inspect, re, os, webbrowser
from .tags import DOC_TAGS
from ..widget import Notebook, Treeview, Menu, Text
from ..style import Color, getimage
from ..util.argspec import argspec
from ..util.docscrape import FunctionDoc
from ..util.common import open_editor, classname
from ..util.config import EXCLUDED_MODULES, BUTTON_RIGHT, BUTTON_RELEASE_RIGHT, INSTALLED_PKGS, BUILTIN_PKGS, DOC_FONT
from .show_functiondoc import show_functiondoc
from .show_textfile import show_textfile
from numpy import ufunc

class Doc(tk.Frame):
	def __init__(self, parent, obj=None, title=None, has_sidebar=None, run_code=None, **kwargs):
		tk.Frame.__init__(self, parent, background=Color.BLACK)
		self.parent = parent
		self.paned_window = ttk.PanedWindow(self, orient="horizontal")
		self.title = title

		frame = tk.Frame(self, background=Color.BACKGROUND)
		self.nav = Text(frame, padx=16, height=2, readonly=True, font=DOC_FONT, cursor="arrow", insertbackground=Color.BACKGROUND, selectbackground=Color.BACKGROUND)
		self.text = Text(frame, readonly=True, has_scrollbar=False, font=DOC_FONT, cursor="arrow", insertbackground=Color.BACKGROUND)

		self.notebook = Notebook(self)
		self.nav_count = 0

		self.tree = Treeview(self, drag=False)

		self.has_sidebar = (has_sidebar or inspect.ismodule(obj) or inspect.isclass(obj))
		self.run_code = run_code
		if self.has_sidebar:
			self.notebook.add("tree", self.tree)
			self.paned_window.add(self.notebook.frame)
			self.nav.pack(side="top", fill="x")

		self.text.pack(side="bottom", fill="both", expand=True)

		self.paned_window.add(frame)
		self.paned_window.pack(side="left", fill="both", expand=True)

		self.menu = Menu(self)
		self.functions = {}
		self.builtins = {}
		self.classes = {}
		self.submodules = {}

		self.parent.bind("<Escape>", lambda event: parent.destroy())
		self.text.bind("<Escape>", lambda event: parent.destroy())

		self.parent.bind("<Key>", self._on_key)
		self.text.bind("<Key>", self._on_key)

		self.tree.bind(BUTTON_RIGHT, self._on_button_right)

		self.tree.tag_bind("submodule", "<ButtonRelease-1>", self._on_button_release_1)
		self.tree.tag_bind("class", "<ButtonRelease-1>", self._on_button_release_1)
		self.tree.bind("<<TreeviewSelect>>", self._on_select)

		for i in ("link_url", "doc_link", "code_sample", "submodule"):
			self.text.tag_bind(i, "<Motion>", lambda event, key=i: self.text._motion(event, key))
			self.text.tag_bind(i, "<Leave>", lambda event, key=i: self.text._leave(event, key))
			self.text.tag_bind(i, "<Button-1>", lambda event, key=i: self._click(event, key))

		for i in ("root", "module_nav"):
			self.nav.tag_bind(i, "<Motion>", lambda event, key=i: self.nav._motion(event, key))
			self.nav.tag_bind(i, "<Leave>", lambda event, key=i: self.nav._leave(event, key))
			self.nav.tag_bind(i, "<Button-1>", lambda event, key=i: self._click_nav(event, key))

		for i in DOC_TAGS:
			self.nav.tag_configure(i, **DOC_TAGS[i])
			self.text.tag_configure(i, **DOC_TAGS[i])

		self.objects = []
		if obj:
			self.show(obj)

	def show(self, obj, clear_nav=True, display_only=False):
		self.obj = obj

		if display_only:
			pass
		elif clear_nav:
			name = self.getname(obj)
			self.nav.delete("1.0", "end")
			self.nav.insert("end", name, ("root", "basepath"))
			self.rootmodule = self.obj
			self.nav_count = 0
			self.objects.clear()
		else:
			name = self.getname(obj).split(".")[-1]
			tag_ranges = self.nav.tag_ranges("subpath")
			if tag_ranges:
				self.nav.delete(*tag_ranges)
			self.nav.insert("end", " > ", ("blue", "nav_count="+str(self.nav_count)))
			self.nav.insert("end", name, ("module_nav", "basepath", "nav_count="+str(len(self.objects))))
			self.objects.append(obj)

		self.clear()

		if isinstance(obj, str) and os.path.isfile(obj):
			content = open(obj).read()
			show_textfile(self.text, content)
			return

		for i in dir(self.obj):
			attr = getattr(self.obj, i)
			if i[0] == "_":
				pass
			elif inspect.isclass(attr):
				self.classes[i] = attr
			elif inspect.ismodule(attr):# and i not in INSTALLED_PKGS + BUILTIN_PKGS + EXCLUDED_MODULES + ["os", "sys"]:
				self.submodules[i] = attr
			elif inspect.isfunction(attr):
				self.functions[i] = attr
			elif callable(attr):
				self.builtins[i] = attr
			else:
				pass

		if self.classes:
			classes = self.tree.insert("", "end", text="classes", open=True)
			for k in self.classes:
				temp = self.tree.insert(classes, "end", k, text=k, tags="class")

		if self.builtins:
			if inspect.isclass(self.obj):
				parent = self.tree.insert("", "end", text="methods", open=True)
			else:
				parent = self.tree.insert("", "end", text="builtins", open=True)
			for j in self.builtins:
				self.tree.insert(parent, "end", j, text=j)

		if self.functions:
			functions = self.tree.insert("", "end", text="functions", open=True)
			for j in self.functions:
				self.tree.insert(functions, "end", j, text=j, open=True)

		if self.submodules:
			for i in self.submodules:
				self.tree.insert("", "end", i, image=getimage(".py"), text="      " + i, tags="submodule")

		if self.has_sidebar:
			if not self.builtins and not self.functions and not self.classes and not self.submodules:
				if self.paned_window.winfo_width() > 1:
					self.paned_window.sashpos(0,0)
				else:
					self.paned_window.bind("<Configure>", lambda event: self.paned_window.sashpos(0,0))
			elif self.paned_window.sashpos(0) < 20:
				self.paned_window.sashpos(0,220)

		self.display_doc(obj)

	def display_doc(self, obj):
		if inspect.ismodule(obj) or inspect.isclass(obj):
			show_textfile(self.text, inspect.getdoc(obj))
			return

		try:
			doc = FunctionDoc(obj)
		except Exception:
			show_textfile(self.text, inspect.getdoc(obj))
			# self.toggle_scrollbar()
			return
		show_functiondoc(self.text, doc, classname(obj))
		# self.toggle_scrollbar()

	# REFACTOR - need a better system for scrollbars in general
	def toggle_scrollbar(self):
		self.doc.delete("1.0", "end")
		if self.text.yview()[1] == 1.0:
			self.text.scrollbar.pack_forget()
		elif not self.text.scrollbar.winfo_ismapped():
			if self.has_sidebar:
				self.text.scrollbar.pack(before=self.nav, side="right", fill="y")
			else:
				self.text.scrollbar.pack(before=self.text, side="right", fill="y")

	def getname(self, obj):
		return self.title if self.title else obj.__name__ if hasattr(obj, "__name__") else obj.__class__.__name__

	def clear(self):
		self.text.delete("1.0", "end")

		self.builtins.clear()
		self.functions.clear()
		self.classes.clear()
		self.submodules.clear()

		for i in self.tree.get_children():
			self.tree.delete(i)

	def _on_select(self, event):
		try:
			key = self.tree.selection()[0]
		except IndexError:
			return
		if not hasattr(self.obj, key): return

		obj = getattr(self.obj, key)
		self.text.delete("1.0", "end")
		tag_ranges = self.nav.tag_ranges("subpath")
		if tag_ranges:
			self.nav.delete(*tag_ranges)
		self.nav.insert("end", " > ", ("blue", "subpath"))
		self.nav.insert("end", key, ("module_nav", "subpath", "nav_count="+str(self.nav_count)))
		self.display_doc(obj)

	def _on_button_release_1(self, event):
		key = self.tree.selection()[0]
		self.show(getattr(self.obj, key), clear_nav=False)

	def _on_button_right(self, event):
		name = self.tree.identify_row(event.y)
		obj = getattr(self.obj, name)

		try:
			file = inspect.getsourcefile(obj)
		except:
			file = None

		if file:
			self.menu.show(event, [{
				"label": "View Source Code",
				"command": lambda: open_editor(file)
			}])

	def _click(self, event, tag):
		if tag == "link_url":
			webbrowser.open(self.text.get(*self.text.hover_range), new=2)
		elif tag == "doc_link":
			doc_link = self.text.get(*self.text.hover_range)
			obj = getattr(__import__(self.obj.__class__.__module__), doc_link)
			self.show(obj)
		elif tag == "code_sample" and self.run_code:
			match = re.findall(r"(>>>|\.\.\.) {0,2}(\t{0,}.*\n)", self.text.get(*self.text.hover_range))
			if not match: return
			for command in match:
				self.run_code(command[1])

	def _click_nav(self, event, tag):
		if tag == "root":
			self.show(self.rootmodule)
		elif tag == "module_nav":
			tags = self.nav.get_tags(*self.nav.hover_range)
			nav_count = None
			for i in tags:
				if i[:10] == "nav_count=":
					nav_count = int(i[10:])
					break
			if nav_count is None: return

			self.nav.delete(self.nav.hover_range[1], "end")
			self.show(self.objects[nav_count], display_only=True)


	def _on_key(self, event):
		ctrl = (event.state & 0x4) != 0
		meta = (event.state & 0x8) != 0
		is_mod = ctrl or meta
		if is_mod and event.char == "w":
			return self.parent.destroy()
		return self.text._on_key(event)
